"""Linear models based on Torch library."""

import logging

from copy import deepcopy
from typing import Callable
from typing import Optional
from typing import Sequence
from typing import Union

import numpy as np
import torch

from scipy import sparse
from torch import nn
from torch import optim

from ...tasks.losses import TorchLossWrapper
from ..utils import SoftmaxClip


logger = logging.getLogger(__name__)
ArrayOrSparseMatrix = Union[np.ndarray, sparse.spmatrix]


def convert_scipy_sparse_to_torch_float(matrix: sparse.spmatrix) -> torch.Tensor:
    """Convert scipy sparse matrix to torch sparse tensor.

    Args:
        matrix: Matrix to convert.

    Returns:
        Matrix in torch.Tensor format.

    """
    matrix = sparse.coo_matrix(matrix, dtype=np.float32)
    np_idx = np.stack([matrix.row, matrix.col], axis=0).astype(np.int64)
    idx = torch.from_numpy(np_idx)
    values = torch.from_numpy(matrix.data)
    sparse_tensor = torch.sparse_coo_tensor(idx, values, size=matrix.shape)

    return sparse_tensor


class CatLinear(nn.Module):
    """Simple linear model to handle numeric and categorical features.

    Args:
        numeric_size: Number of numeric features.
        embed_sizes: Embedding sizes.
        output_size: Size of output layer.

    """

    def __init__(
        self,
        numeric_size: int = 0,
        embed_sizes: Sequence[int] = (),
        output_size: int = 1,
    ):
        super().__init__()
        self.bias = nn.Parameter(torch.zeros(output_size))
        # add numeric if it is defined
        self.linear = None
        if numeric_size > 0:
            self.linear = nn.Linear(in_features=numeric_size, out_features=output_size, bias=False)
            nn.init.zeros_(self.linear.weight)

        # add categories if it is defined
        self.cat_params = None
        if len(embed_sizes) > 0:
            self.cat_params = nn.Parameter(torch.zeros(sum(embed_sizes), output_size))
            self.embed_idx = torch.LongTensor(embed_sizes).cumsum(dim=0) - torch.LongTensor(embed_sizes)
        self.final_act = nn.Identity()

    def forward(
        self,
        numbers: Optional[torch.Tensor] = None,
        categories: Optional[torch.Tensor] = None,
    ):
        """Forward-pass.

        Args:
            numbers: Input numeric features.
            categories: Input categorical features.

        Returns:
            Linear prediction.

        """
        x = self.bias

        if self.linear is not None:
            x = x + self.linear(numbers)

        if self.cat_params is not None:
            x = x + self.cat_params[categories + self.embed_idx].sum(dim=1)

        return x

    def predict(
        self,
        numbers: Optional[torch.Tensor] = None,
        categories: Optional[torch.Tensor] = None,
    ):
        """Inference phase.

        Args:
            numbers: Numeric data.
            categories: Categorical data.

        Returns:
            Predicted logits/targets for cls/other tasks.

        """
        x = self.forward(numbers, categories)
        x = self.final_act(x)
        return x


class CatLogisticRegression(CatLinear):
    """Realisation of torch-based logistic regression."""

    def __init__(self, numeric_size: int, embed_sizes: Sequence[int] = (), output_size: int = 1):
        super().__init__(numeric_size, embed_sizes=embed_sizes, output_size=output_size)
        self.final_act = nn.Sigmoid()


class CatRegression(CatLinear):
    """Realisation of torch-based linear regreession."""

    def __init__(self, numeric_size: int, embed_sizes: Sequence[int] = (), output_size: int = 1):
        super().__init__(numeric_size, embed_sizes=embed_sizes, output_size=output_size)


class CatMulticlass(CatLinear):
    """Realisation of multi-class linear classifier."""

    def __init__(self, numeric_size: int, embed_sizes: Sequence[int] = (), output_size: int = 1):
        super().__init__(numeric_size, embed_sizes=embed_sizes, output_size=output_size)
        self.final_act = SoftmaxClip(dim=1)


class TorchBasedLinearEstimator:
    """Linear model based on torch L-BFGS solver.

    Accepts Numeric + Label Encoded categories or Numeric sparse input.

    Args:
        data_size: Not used.
        categorical_idx: Indices of categorical features.
        embed_sizes: Categorical embedding sizes.
        output_size: Size of output layer.
        cs: Regularization coefficients.
        max_iter: Maximum iterations of L-BFGS.
        tol: Tolerance for the stopping criteria.
        early_stopping: Maximum rounds without improving.
        loss: Loss function. Format: loss(preds, true) -> loss_arr, assume ```reduction='none'```.
        metric: Metric function. Format: metric(y_true, y_preds, sample_weight = None) -> float (greater_is_better).

    """

    def __init__(
        self,
        data_size: int,
        categorical_idx: Sequence[int] = (),
        embed_sizes: Sequence[int] = (),
        output_size: int = 1,
        cs: Sequence[float] = (
            0.00001,
            0.00005,
            0.0001,
            0.0005,
            0.001,
            0.005,
            0.01,
            0.05,
            0.1,
            0.5,
            1.0,
            2.0,
            5.0,
            7.0,
            10.0,
            20.0,
        ),
        max_iter: int = 1000,
        tol: float = 1e-5,
        early_stopping: int = 2,
        loss=Optional[Callable],
        metric=Optional[Callable],
    ):
        self.data_size = data_size
        self.categorical_idx = categorical_idx
        self.embed_sizes = embed_sizes
        self.output_size = output_size

        assert all([x > 0 for x in cs]), "All Cs should be greater than 0"

        self.cs = cs
        self.max_iter = max_iter
        self.tol = tol
        self.early_stopping = early_stopping
        self.loss = loss  # loss(preds, true) -> loss_arr, assume reduction='none'
        self.metric = metric  # metric(y_true, y_preds, sample_weight = None) -> float (greater_is_better)

    def _prepare_data(self, data: ArrayOrSparseMatrix):
        """Prepare data based on input type.

        Args:
            data: Data to prepare.

        Returns:
            Tuple (numeric_features, cat_features).

        """
        if sparse.issparse(data):
            return self._prepare_data_sparse(data)

        return self._prepare_data_dense(data)

    def _prepare_data_sparse(self, data: sparse.spmatrix):
        """Prepare sparse matrix.

        Only supports numeric features.

        Args:
            data: data to prepare.

        Returns:
            Tuple (numeric_features, `None`).

        """
        assert len(self.categorical_idx) == 0, "Support only numeric with sparse matrix"
        data = convert_scipy_sparse_to_torch_float(data)
        return data, None

    def _prepare_data_dense(self, data: np.ndarray):
        """Prepare dense matrix.

        Split categorical and numeric features.

        Args:
            data: data to prepare.

        Returns:
            Tuple (numeric_features, cat_features).

        """
        if 0 < len(self.categorical_idx) < data.shape[1]:
            data_cat = torch.from_numpy(data[:, self.categorical_idx].astype(np.int64))
            data = torch.from_numpy(data[:, np.setdiff1d(np.arange(data.shape[1]), self.categorical_idx)])
            return data, data_cat

        elif len(self.categorical_idx) == 0:
            data = torch.from_numpy(data)
            return data, None

        else:
            data_cat = torch.from_numpy(data.astype(np.int64))
            return None, data_cat

    def _optimize(
        self,
        data: torch.Tensor,
        data_cat: Optional[torch.Tensor],
        y: torch.Tensor = None,
        weights: Optional[torch.Tensor] = None,
        c: float = 1,
    ):
        """Optimize single model.

        Args:
            data: Numeric data to train.
            data_cat: Categorical data to train.
            y: Target values.
            weights: Item weights.
            c: Regularization coefficient.

        """
        self.model.train()
        opt = optim.LBFGS(
            self.model.parameters(),
            lr=0.1,
            max_iter=self.max_iter,
            tolerance_change=self.tol,
            tolerance_grad=self.tol,
            line_search_fn="strong_wolfe",
        )

        # keep history
        results = []

        def closure():
            opt.zero_grad()

            output = self.model(data, data_cat)
            loss = self._loss_fn(y, output, weights, c)
            if loss.requires_grad:
                loss.backward()
            results.append(loss.item())
            return loss

        opt.step(closure)

    def _loss_fn(
        self,
        y_true: torch.Tensor,
        y_pred: torch.Tensor,
        weights: Optional[torch.Tensor],
        c: float,
    ) -> torch.Tensor:
        """Weighted loss_fn wrapper.

        Args:
            y_true: True target values.
            y_pred: Predicted target values.
            weights: Item weights.
            c: Regularization coefficients.

        Returns:
            Loss+Regularization value.

        """
        # weighted loss
        loss = self.loss(y_true, y_pred, sample_weight=weights)

        n = y_true.shape[0]
        if weights is not None:
            n = weights.sum()

        all_params = torch.cat([y.view(-1) for (x, y) in self.model.named_parameters() if x != "bias"])

        penalty = torch.norm(all_params, 2).pow(2) / 2 / n

        return loss + 0.5 * penalty / c

    def fit(
        self,
        data: np.ndarray,
        y: np.ndarray,
        weights: Optional[np.ndarray] = None,
        data_val: Optional[np.ndarray] = None,
        y_val: Optional[np.ndarray] = None,
        weights_val: Optional[np.ndarray] = None,
    ):
        """Fit method.

        Args:
            data: Data to train.
            y: Train target values.
            weights: Train items weights.
            data_val: Data to validate.
            y_val: Valid target values.
            weights_val: Validation item weights.

        Returns:
            self.

        """
        assert self.model is not None, "Model should be defined"
        data, data_cat = self._prepare_data(data)
        if len(y.shape) == 1:
            y = y[:, np.newaxis]
        y = torch.from_numpy(y.astype(np.float32))
        if weights is not None:
            weights = torch.from_numpy(weights.astype(np.float32))

        if data_val is None and y_val is None:
            logger.info2("Validation data should be defined. No validation will be performed and C = 1 will be used")
            self._optimize(data, data_cat, y, weights, 1.0)

            return self

        data_val, data_val_cat = self._prepare_data(data_val)

        best_score = -np.inf
        best_model = None
        es = 0

        for c in self.cs:
            self._optimize(data, data_cat, y, weights, c)

            val_pred = self._score(data_val, data_val_cat)
            score = self.metric(y_val, val_pred, weights_val)
            logger.info3("Linear model: C = {0} score = {1}".format(c, score))
            if score > best_score:
                best_score = score
                best_model = deepcopy(self.model)
                es = 0
            else:
                es += 1

            if es >= self.early_stopping:
                break

        self.model = best_model

        return self

    def _score(self, data: np.ndarray, data_cat: Optional[np.ndarray]) -> np.ndarray:
        """Get predicts to evaluate performance of model.

        Args:
            data: Numeric data.
            data_cat: Categorical data.

        Returns:
            Predicted target values.

        """
        with torch.set_grad_enabled(False):
            self.model.eval()
            preds = self.model.predict(data, data_cat).numpy()

        return preds

    def predict(self, data: np.ndarray) -> np.ndarray:
        """Inference phase.

        Args:
            data: Data to test.

        Returns:
            Predicted target values.

        """
        data, data_cat = self._prepare_data(data)

        return self._score(data, data_cat)


class TorchBasedLogisticRegression(TorchBasedLinearEstimator):
    """Linear binary classifier.

    Args:
        data_size: not used.
        categorical_idx: indices of categorical features.
        embed_sizes: categorical embedding sizes.
        output_size: size of output layer.
        cs: regularization coefficients.
        multilabel: multilabel or not.
        max_iter: maximum iterations of L-BFGS.
        tol: the tolerance for the stopping criteria.
        early_stopping: maximum rounds without improving.
        loss: loss function. Format: loss(preds, true) -> loss_arr, assume reduction='none'.
        metric: metric function. Format: metric(y_true, y_preds, sample_weight = None) -> float (greater_is_better).

    """

    def __init__(
        self,
        data_size: int,
        categorical_idx: Sequence[int] = (),
        embed_sizes: Sequence[int] = (),
        output_size: int = 1,
        cs: Sequence[float] = (
            0.00001,
            0.00005,
            0.0001,
            0.0005,
            0.001,
            0.005,
            0.01,
            0.05,
            0.1,
            0.5,
            1.0,
            2.0,
            5.0,
            7.0,
            10.0,
            20.0,
        ),
        multilabel: bool = False,
        max_iter: int = 1000,
        tol: float = 1e-4,
        early_stopping: int = 2,
        loss=Optional[Callable],
        metric=Optional[Callable],
    ):
        if multilabel:
            _loss = nn.BCEWithLogitsLoss
            _model = CatLogisticRegression
            self._binary = False
        elif output_size == 1:
            _loss = nn.BCEWithLogitsLoss
            _model = CatLogisticRegression
            self._binary = True
        else:
            _loss = nn.CrossEntropyLoss
            _model = CatMulticlass
            self._binary = False
        if loss is None:
            loss = TorchLossWrapper(_loss)

        super().__init__(
            data_size,
            categorical_idx,
            embed_sizes,
            output_size,
            cs,
            max_iter,
            tol,
            early_stopping,
            loss,
            metric,
        )
        self.model = _model(
            self.data_size - len(self.categorical_idx),
            self.embed_sizes,
            self.output_size,
        )

    def predict(self, data: np.ndarray) -> np.ndarray:
        """Inference phase.

        Args:
            data: data to test.

        Returns:
            predicted target values.

        """
        pred = super().predict(data)
        if self._binary:
            pred = pred[:, 0]
        return pred


class TorchBasedLinearRegression(TorchBasedLinearEstimator):
    """Torch-based linear regressor optimized by L-BFGS.

    Args:
        data_size: used only for super function.
        categorical_idx: indices of categorical features.
        embed_sizes: categorical embedding sizes
        output_size: size of output layer.
        cs: regularization coefficients.
        max_iter: maximum iterations of L-BFGS.
        tol: the tolerance for the stopping criteria.
        early_stopping: maximum rounds without improving.
        loss: loss function. Format: loss(preds, true) -> loss_arr, assume reduction='none'.
        metric: metric function. Format: metric(y_true, y_preds, sample_weight = None) -> float (greater_is_better).

    """

    def __init__(
        self,
        data_size: int,
        categorical_idx: Sequence[int] = (),
        embed_sizes: Sequence[int] = (),
        output_size: int = 1,
        cs: Sequence[float] = (
            0.00001,
            0.00005,
            0.0001,
            0.0005,
            0.001,
            0.005,
            0.01,
            0.05,
            0.1,
            0.5,
            1.0,
            2.0,
            5.0,
            7.0,
            10.0,
            20.0,
        ),
        max_iter: int = 1000,
        tol: float = 1e-4,
        early_stopping: int = 2,
        loss=Optional[Callable],
        metric=Optional[Callable],
    ):
        if loss is None:
            loss = TorchLossWrapper(nn.MSELoss)
        super().__init__(
            data_size,
            categorical_idx,
            embed_sizes,
            output_size,
            cs,
            max_iter,
            tol,
            early_stopping,
            loss,
            metric,
        )
        self.model = CatRegression(
            self.data_size - len(self.categorical_idx),
            self.embed_sizes,
            self.output_size,
        )

    def predict(self, data: np.ndarray) -> np.ndarray:
        """Inference phase.

        Args:
            data: data to test.

        Returns:
            predicted target values.

        """
        if self.output_size == 1:
            return super().predict(data)[:, 0]
        else:
            return super().predict(data)



"""Wrapped LightGBM for tabular datasets."""

import logging

from contextlib import redirect_stdout
from copy import copy
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Tuple

import xgboost
import numpy as np

from pandas import Series

from ..pipelines.selection.base import ImportanceEstimator
from ..utils.logging import LoggerStream
from ..validation.base import TrainValidIterator
from .base import TabularDataset
from .base import TabularMLAlgo
from .tuning.base import Uniform, Choice


logger = logging.getLogger(__name__)


class BoostXGB(TabularMLAlgo, ImportanceEstimator):
    """Gradient boosting on decision trees from XGBoost library.

    default_params: All available parameters listed in XGBoost documentation:

        - https://xgboost.readthedocs.io/en/stable/parameter.html

    freeze_defaults:

        - ``True`` :  params may be rewritten depending on dataset.
        - ``False``:  params may be changed only manually or with tuning.

    timer: :class:`~lightautoml.utils.timer.Timer` instance or ``None``.

    """

    _name: str = "XGBoost"

    _default_params = {
        # "task": "train",
        # "learning_rate": 0.05, #eta
        # "max_leaves": 128,
        # "colsample_bytree": 0.7,
        # "subsample": 0.7 ,
        # # "bagging_freq": 1,
        # "max_depth": -1,
        # "verbosity": -1,
        # "reg_alpha": 1,
        # "reg_lambda": 0.0,
        # "gamma": 0.0,
        # # "zero_as_missing": False,
        # "nthread": 4,
        # "max_bin": 255,
        # # "min_data_in_bin": 3,
        "n_estimators": 3000,
        "early_stopping_rounds": 100,
        "seed": 42,
    }

    def _infer_params(
        self,
    ) -> Tuple[dict, int, int, int, Optional[Callable], Optional[Callable]]:
        """Infer all parameters in lightgbm format.

        Returns:
            Tuple (params, num_trees, early_stopping_rounds, verbose_eval, fobj, feval).
            About parameters: https://lightgbm.readthedocs.io/en/latest/_modules/lightgbm/engine.html

        """
        # TODO: Check how it works with custom tasks
        params = copy(self.params)
        early_stopping_rounds = params.pop("early_stopping_rounds")
        num_trees = params.pop("n_estimators")

        verbose_eval = 100

        # get objective params
        loss = self.task.losses["xgb"]
        params["objective"] = loss.fobj_name
        fobj = loss.fobj

        # # get metric params
        params["eval_metric"] = loss.metric_name
        feval = loss.feval

        params["num_class"] = self.n_classes
        # add loss and tasks params if defined
        params = {**params, **loss.fobj_params, **loss.metric_params}

        return params, num_trees, early_stopping_rounds, verbose_eval, fobj, feval

    def init_params_on_input(self, train_valid_iterator: TrainValidIterator) -> dict:
        """Get model parameters depending on dataset parameters.

        Args:
            train_valid_iterator: Classic cv-iterator.

        Returns:
            Parameters of model.

        """
        # TODO: use features_num
        # features_num = len(train_valid_iterator.features())

        # rows_num = len(train_valid_iterator.train)
        # task = train_valid_iterator.train.task.name

        suggested_params = copy(self.default_params)

        if self.freeze_defaults:
            # if user change defaults manually - keep it
            return suggested_params

        # if task == "reg":
        #     suggested_params = {
        #         "learning_rate": 0.05,
        #         "num_leaves": 32,
        #         "feature_fraction": 0.9,
        #         "bagging_fraction": 0.9,
        #     }

        # if rows_num <= 10000:
        #     init_lr = 0.01
        #     ntrees = 3000
        #     es = 200

        # elif rows_num <= 20000:
        #     init_lr = 0.02
        #     ntrees = 3000
        #     es = 200

        # elif rows_num <= 100000:
        #     init_lr = 0.03
        #     ntrees = 1200
        #     es = 200
        # elif rows_num <= 300000:
        #     init_lr = 0.04
        #     ntrees = 2000
        #     es = 100
        # else:
        #     init_lr = 0.05
        #     ntrees = 2000
        #     es = 100

        # if rows_num > 300000:
        #     suggested_params["num_leaves"] = 128 if task == "reg" else 244
        # elif rows_num > 100000:
        #     suggested_params["num_leaves"] = 64 if task == "reg" else 128
        # elif rows_num > 50000:
        #     suggested_params["num_leaves"] = 32 if task == "reg" else 64
        #     # params['reg_alpha'] = 1 if task == 'reg' else 0.5
        # elif rows_num > 20000:
        #     suggested_params["num_leaves"] = 32 if task == "reg" else 32
        #     suggested_params["reg_alpha"] = 0.5 if task == "reg" else 0.0
        # elif rows_num > 10000:
        #     suggested_params["num_leaves"] = 32 if task == "reg" else 64
        #     suggested_params["reg_alpha"] = 0.5 if task == "reg" else 0.2
        # elif rows_num > 5000:
        #     suggested_params["num_leaves"] = 24 if task == "reg" else 32
        #     suggested_params["reg_alpha"] = 0.5 if task == "reg" else 0.5
        # else:
        #     suggested_params["num_leaves"] = 16 if task == "reg" else 16
        #     suggested_params["reg_alpha"] = 1 if task == "reg" else 1

        # suggested_params["learning_rate"] = init_lr
        # suggested_params["num_trees"] = ntrees
        # suggested_params["early_stopping_rounds"] = es

        return suggested_params

    def _get_default_search_spaces(self, suggested_params: Dict, estimated_n_trials: int) -> Dict:
        """Sample hyperparameters from suggested.

        Args:
            suggested_params: Dict with parameters.
            estimated_n_trials: Maximum number of hyperparameter estimations.

        Returns:
            dict with sampled hyperparameters.

        """
        optimization_search_space = {}

        optimization_search_space["colsample_bytree"] = Choice(options=[0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0])
        optimization_search_space["subsample"] = Choice(options=[0.4, 0.5, 0.6, 0.7, 0.8, 1.0])
        optimization_search_space["max_depth"] = Choice(options=[5, 7, 9, 11, 13, 15, 17])
        optimization_search_space["learning_rate"] = Choice(options=[0.008, 0.01, 0.012, 0.014, 0.016, 0.018, 0.02])

        if estimated_n_trials > 30:
            optimization_search_space["min_child_weight"] = Uniform(low=1, high=300, q=1, log=False)

            optimization_search_space["reg_alpha"] = Uniform(
                low=1e-3,
                high=10.0,
                log=True,
            )
            optimization_search_space["reg_lambda"] = Uniform(
                low=1e-3,
                high=10.0,
                log=True,
            )

        return optimization_search_space

    def fit_predict_single_fold(
        self, train: TabularDataset, valid: TabularDataset
    ) -> Tuple[xgboost.Booster, np.ndarray]:
        """Implements training and prediction on single fold.

        Args:
            train: Train Dataset.
            valid: Validation Dataset.

        Returns:
            Tuple (model, predicted_values)

        """
        (
            params,
            num_trees,
            early_stopping_rounds,
            verbose_eval,
            fobj,
            feval,
        ) = self._infer_params()

        train_target, train_weight = self.task.losses["xgb"].fw_func(train.target, train.weights)
        valid_target, valid_weight = self.task.losses["xgb"].fw_func(valid.target, valid.weights)

        dtrain = xgboost.DMatrix(train.data, label=train_target, weight=train_weight)
        dval = xgboost.DMatrix(valid.data, label=valid_target, weight=valid_weight)

        with redirect_stdout(LoggerStream(logger, verbose_eval=100)):
            # maximize=None, evals_result=None, xgb_model=None, callbacks=None, custom_metric=None)
            model = xgboost.train(
                params=params,
                dtrain=dtrain,
                num_boost_round=num_trees,
                evals=[(dval, "valid"), (dtrain, "train")],
                obj=fobj,
                feval=feval,
                early_stopping_rounds=early_stopping_rounds,
                verbose_eval=verbose_eval,
            )

        val_pred = model.predict(data=dval)
        val_pred = self.task.losses["xgb"].bw_func(val_pred)

        return model, val_pred

    def predict_single_fold(self, model: xgboost.Booster, dataset: TabularDataset) -> np.ndarray:
        """Predict target values for dataset.

        Args:
            model: Xgboost object.
            dataset: Test Dataset.

        Returns:
            Predicted target values.

        """
        pred = self.task.losses["xgb"].bw_func(model.predict(xgboost.DMatrix(dataset.data)))

        return pred

    def get_features_score(self) -> Series:
        """Computes feature importance as mean values of feature importance provided by lightgbm per all models.

        Returns:
            Series with feature importances.

        """
        imp = 0
        for model in self.models:
            imp = imp + model.feature_importance(importance_type="gain")

        imp = imp / len(self.models)

        return Series(imp, index=self.features).sort_values(ascending=False)

    def fit(self, train_valid: TrainValidIterator):
        """Just to be compatible with :class:`~lightautoml.pipelines.selection.base.ImportanceEstimator`.

        Args:
            train_valid: Classic cv-iterator.

        """
        self.fit_predict(train_valid)



"""Classes to implement hyperparameter tuning using Optuna."""

import logging

from copy import copy
from copy import deepcopy
from typing import Callable
from typing import Optional
from typing import Tuple
from typing import TypeVar
from typing import Union

import optuna
from tqdm import tqdm

from ...dataset.base import LAMLDataset
from ..base import MLAlgo
from .base import Choice
from .base import ParamsTuner
from .base import Uniform
from ...validation.base import HoldoutIterator
from ...validation.base import TrainValidIterator
from ...utils.logging import get_stdout_level


logger = logging.getLogger(__name__)
optuna.logging.enable_propagation()
optuna.logging.disable_default_handler()
optuna.logging.set_verbosity(optuna.logging.DEBUG)

TunableAlgo = TypeVar("TunableAlgo", bound=MLAlgo)


class ChoiceWrapOptuna:
    """TODO."""

    def __init__(self, choice) -> None:
        self.choice = choice

    def __call__(self, name, trial):
        """_summary_.

        Args:
            name (_type_): _description_
            trial (_type_): _description_

        Returns:
            _type_: _description_
        """
        return trial.suggest_categorical(name=name, choices=self.choice.options)


class UniformWrapOptuna:
    """TODO."""

    def __init__(self, choice) -> None:
        self.choice = choice

    def __call__(self, name, trial):
        """_summary_.

        Args:
            name (_type_): _description_
            trial (_type_): _description_

        Returns:
            _type_: _description_
        """
        if (self.choice.q is not None) and float(self.choice.q).is_integer() and (self.choice.q == 1):
            result = trial.suggest_int(name=name, low=self.choice.low, high=self.choice.high)
        else:
            result = trial.suggest_float(
                name=name, low=self.choice.low, high=self.choice.high, step=self.choice.q, log=self.choice.log
            )

        return result


OPTUNA_DISTRIBUTIONS_MAP = {Choice: ChoiceWrapOptuna, Uniform: UniformWrapOptuna}


class OptunaTuner(ParamsTuner):
    """Wrapper for optuna tuner.

    Args:
        timeout: Maximum learning time.
        n_trials: Maximum number of trials.
        direction: Direction of optimization.
            Set ``minimize`` for minimization
            and ``maximize`` for maximization.
        fit_on_holdout: Will be used holdout cv-iterator.
        random_state: Seed for optuna sampler.
        fail_tolerance: the maximum allowed percentage of failed tuner trials. Exception will be thrown after crossing the threshold value.

    """

    _name: str = "OptunaTuner"

    study: optuna.study.Study = None
    estimated_n_trials: int = None
    mean_trial_time: Optional[int] = None

    def __init__(
        # TODO: For now, metric is designed to be greater is better. Change maximize param after metric refactor if needed
        self,
        timeout: Optional[int] = 1000,
        n_trials: Optional[int] = 100,
        direction: Optional[str] = "maximize",
        fit_on_holdout: bool = True,
        random_state: int = 42,
        fail_tolerance: float = 0.5,
    ):
        self.timeout = timeout
        self.n_trials = n_trials
        self.estimated_n_trials = n_trials
        self.direction = direction
        self._fit_on_holdout = fit_on_holdout
        self.random_state = random_state
        self.fail_tolerance = fail_tolerance

    def _upd_timeout(self, timeout):
        self.timeout = min(self.timeout, timeout)

    def fit(
        self,
        ml_algo: TunableAlgo,
        train_valid_iterator: Optional[TrainValidIterator] = None,
    ) -> Tuple[Optional[TunableAlgo], Optional[LAMLDataset]]:
        """Tune model.

        Args:
            ml_algo: Algo that is tuned.
            train_valid_iterator: Classic cv-iterator.

        Returns:
            Tuple (None, None) if an optuna exception raised
            or ``fit_on_holdout=True`` and ``train_valid_iterator`` is
            not :class:`~lightautoml.validation.base.HoldoutIterator`.
            Tuple (MlALgo, preds_ds) otherwise.

        """
        assert not ml_algo.is_fitted, "Fitted algo cannot be tuned."
        # optuna.logging.set_verbosity(logger.getEffectiveLevel())
        # upd timeout according to ml_algo timer
        estimated_tuning_time = ml_algo.timer.estimate_tuner_time(len(train_valid_iterator))
        if estimated_tuning_time:
            # TODO: Check for minimal runtime!
            estimated_tuning_time = max(estimated_tuning_time, 1)
            self._upd_timeout(estimated_tuning_time)

        logger.info(
            f"Start hyperparameters optimization for \x1b[1m{ml_algo._name}\x1b[0m ... Time budget is {self.timeout:.2f} secs"
        )

        metric_name = train_valid_iterator.train.task.get_dataset_metric().name
        ml_algo = deepcopy(ml_algo)

        flg_new_iterator = False
        if self._fit_on_holdout and not isinstance(train_valid_iterator, HoldoutIterator):
            train_valid_iterator = train_valid_iterator.convert_to_holdout_iterator()
            flg_new_iterator = True

        # TODO: Check if time estimation will be ok with multiprocessing
        def update_trial_time(study: optuna.study.Study, trial: optuna.trial.FrozenTrial):
            """Callback for number of iteration with time cut-off.

            Args:
                study: Optuna study object.
                trial: Optuna trial object.

            """
            ml_algo.mean_trial_time = study.trials_dataframe()["duration"].mean().total_seconds()
            self.estimated_n_trials = min(self.n_trials, self.timeout // ml_algo.mean_trial_time)

            logger.info3(
                f"\x1b[1mTrial {len(study.trials)}\x1b[0m with hyperparameters {trial.params} scored {trial.value} in {trial.duration}"
            )

        def check_fail_tolerance(study: optuna.study.Study, trial: optuna.trial.FrozenTrial):
            df = study.trials_dataframe()

            if (self.estimated_n_trials > 0) and (
                df[df["state"] == "FAIL"].shape[0] / self.estimated_n_trials > self.fail_tolerance
            ):
                raise Exception(
                    f"Too much trials was failed ({df[df['state'] == 'FAIL'].shape[0]} of {df.shape[0]}). Check the model or search space for it."
                )

        try:
            # Custom progress bar
            def custom_progress_bar(study: optuna.study.Study, trial: optuna.trial.FrozenTrial):
                best_trial = study.best_trial
                progress_bar.set_postfix(best_trial=best_trial.number, best_value=best_trial.value)
                progress_bar.update(1)

            # Initialize progress bar
            if get_stdout_level() in [logging.INFO, logging.INFO2]:
                progress_bar = tqdm(total=self.n_trials, desc="Optimization Progress")

            sampler = optuna.samplers.TPESampler(seed=self.random_state)
            self.study = optuna.create_study(direction=self.direction, sampler=sampler)

            self.study.optimize(
                func=self._get_objective(
                    ml_algo=ml_algo,
                    estimated_n_trials=self.estimated_n_trials,
                    train_valid_iterator=train_valid_iterator,
                ),
                n_trials=self.n_trials,
                timeout=self.timeout,
                callbacks=(
                    [update_trial_time, check_fail_tolerance, custom_progress_bar]
                    if get_stdout_level() in [logging.INFO, logging.INFO2]
                    else [update_trial_time, check_fail_tolerance]
                ),
                catch=[Exception],
            )

            # Close the progress bar if it was initialized
            if get_stdout_level() in [logging.INFO, logging.INFO2]:
                progress_bar.close()

            # need to update best params here
            self._best_params = self.study.best_params
            ml_algo.params = self._best_params

            logger.info(f"Hyperparameters optimization for \x1b[1m{ml_algo._name}\x1b[0m completed")
            logger.info2(
                f"The set of hyperparameters \x1b[1m{self._best_params}\x1b[0m\n achieve {self.study.best_value:.4f} {metric_name}"
            )

            if flg_new_iterator:
                # if tuner was fitted on holdout set we dont need to save train results
                return None, None

            preds_ds = ml_algo.fit_predict(train_valid_iterator)

            return ml_algo, preds_ds
        except optuna.exceptions.OptunaError:
            return None, None

    def _get_objective(
        self,
        ml_algo: TunableAlgo,
        estimated_n_trials: int,
        train_valid_iterator: TrainValidIterator,
    ) -> Callable[[optuna.trial.Trial], Union[float, int]]:
        """Get objective.

        Args:
            ml_algo: Tunable algorithm.
            estimated_n_trials: Maximum number of hyperparameter estimations.
            train_valid_iterator: Used for getting parameters
                depending on dataset.

        Returns:
            Callable objective.

        """
        assert isinstance(ml_algo, MLAlgo)

        def objective(trial: optuna.trial.Trial) -> float:
            _ml_algo = deepcopy(ml_algo)

            optimization_search_space = _ml_algo.optimization_search_space

            if not optimization_search_space:
                optimization_search_space = _ml_algo._get_default_search_spaces(
                    suggested_params=_ml_algo.init_params_on_input(train_valid_iterator),
                    estimated_n_trials=estimated_n_trials,
                )

            if callable(optimization_search_space):
                _ml_algo.params = optimization_search_space(
                    trial=trial,
                    estimated_n_trials=estimated_n_trials,
                    suggested_params=_ml_algo.init_params_on_input(train_valid_iterator),
                )
            else:
                _ml_algo.params = self._sample(
                    trial=trial,
                    optimization_search_space=optimization_search_space,
                    suggested_params=_ml_algo.init_params_on_input(train_valid_iterator),
                )

            output_dataset = _ml_algo.fit_predict(train_valid_iterator=train_valid_iterator)

            return _ml_algo.score(output_dataset)

        return objective

    def _sample(
        self,
        optimization_search_space,
        trial: optuna.trial.Trial,
        suggested_params: dict,
    ) -> dict:
        # logger.info3(f'Suggested parameters: {suggested_params}')

        trial_values = copy(suggested_params)

        for parameter_name, search_space in optimization_search_space.items():
            not_supported = True
            for key_class in OPTUNA_DISTRIBUTIONS_MAP:
                if isinstance(search_space, key_class):
                    wrapped_search_space = OPTUNA_DISTRIBUTIONS_MAP[key_class](search_space)
                    trial_values[parameter_name] = wrapped_search_space(
                        name=parameter_name,
                        trial=trial,
                    )
                    not_supported = False
            if not_supported:
                raise ValueError(f"Optuna does not support distribution {search_space}")

        return trial_values

    def plot(self):
        """Plot optimization history of all trials in a study."""
        return optuna.visualization.plot_optimization_history(self.study)



import calendar
import logging

import matplotlib.pyplot as plt
import numpy as np
from copy import copy
import pandas as pd
import seaborn as sns
from ...dataset.utils import concatenate


logger = logging.getLogger(__name__)


def calc_one_feat_imp(iters, level, feat, model, data, norm_score, target, metric, silent):
    # data is a LAMLDataset
    initial_col = copy(data[:, feat].data)
    new_col = np.random.permutation(initial_col)
    data[feat] = new_col
    dataset = copy(data)

    # iterate through higher levels
    for n, level in enumerate(model.levels[level:], level + 1):
        level_predictions = []
        for _n, ml_pipe in enumerate(level):
            level_predictions.append(ml_pipe.predict(dataset))

        if n != len(model.levels):

            level_predictions = concatenate(level_predictions)

            if model.skip_conn:

                try:
                    # convert to initial dataset type
                    level_predictions = dataset.from_dataset(level_predictions)
                except TypeError:
                    raise TypeError("Can not convert prediction dataset type to input features. Set skip_conn=False")
                dataset = concatenate([level_predictions, dataset])
            else:
                dataset = level_predictions

    preds = model.blender.predict(level_predictions)
    data[feat] = initial_col
    preds.target = target.values
    new_score = metric(preds)

    if not silent:
        logger.info3("{}/{} Calculated score for {}: {:.7f}".format(iters[0], iters[1], feat, norm_score - new_score))
    return feat, norm_score - new_score


def calc_feats_permutation_imps(model, used_feats, data, target, metric, silent=False):
    n_used_feats = len(used_feats)
    if not silent:
        logger.info3("LightAutoML ts master used {} feats".format(n_used_feats))
    data = data.reset_index(drop=True)
    preds = model.predict(data)
    target_col = data[target]
    preds.target = target_col.values
    norm_score = metric(preds)
    feat_imp = []

    # construct levels of used feats
    used_feats_leveled = {}

    # input data features
    initial_feats = data.drop(columns=target).columns.values
    used_feats_leveled[0] = initial_feats

    # construct levels of stacking feats
    # only consider features not in initial_feats
    for feat in list(set(used_feats) - set(initial_feats)):
        # stacking feat looks like 'Lvl_0_Pipe_1_Mod_2_CatBoost_prediction_1'
        level = feat.split("_")[1]
        level = int(level) + 1

        arr = used_feats_leveled.get(level, [])
        arr.append(feat)
        used_feats_leveled[level] = arr

    # convert holdout data to LAMLDataset
    data = model.reader.read(data, add_array_attrs=False)
    used_feats_leveled[0] = [feature for feature in data.features if feature not in target]

    # iterate through all the levels
    for level in sorted(used_feats_leveled.keys()):
        # compute importances for features
        for it, f in enumerate(used_feats_leveled[level]):
            feat_imp.append(
                calc_one_feat_imp(
                    (it + 1, n_used_feats),
                    level,
                    f,
                    model,
                    data,
                    norm_score,
                    target_col,
                    metric,
                    silent,
                )
            )
        # update holdout data with stacking features
        if level != len(model.levels) - 1:
            level_predictions = []
            for _n, ml_pipe in enumerate(model.levels[level]):
                level_predictions.append(ml_pipe.predict(data))

            level_predictions = concatenate(level_predictions)
            if model.skip_conn:
                try:
                    # convert to initial dataset type
                    level_predictions = data.from_dataset(level_predictions)
                except TypeError:
                    raise TypeError("Can not convert prediction dataset type to input features. Set skip_conn=False")
                data = concatenate([level_predictions, data])
            else:
                data = level_predictions

    feat_imp = pd.DataFrame(feat_imp, columns=["Feature", "Importance"])
    feat_imp = feat_imp.sort_values("Importance", ascending=False).reset_index(drop=True)
    return feat_imp


def change_datetime(feature_datetime, key, value):
    assert key in ["year", "month", "dayofweek"]
    MAX_DAY = {
        1: 31,
        2: 28,
        3: 31,
        4: 30,
        5: 31,
        6: 30,
        7: 31,
        8: 31,
        9: 30,
        10: 31,
        11: 30,
        12: 31,
    }
    changed = []
    if key == "year":
        year = value
        for i in feature_datetime:
            if i.month == 2 and i.day == 29 and not calendar.isleap(year):
                i -= pd.Timedelta(1, "d")
            changed.append(i.replace(year=year))
    if key == "month":
        month = value
        assert month in np.arange(1, 13)
        for i in feature_datetime:
            if i.day > MAX_DAY[month]:
                i -= pd.Timedelta(i.day - MAX_DAY[month], "d")
                if month == 2 and i.day == 28 and calendar.isleap(i.year):
                    i += pd.Timedelta(1, "d")
            changed.append(i.replace(month=month))
    if key == "dayofweek":
        dayofweek = value
        assert value in np.arange(7)
        for i in feature_datetime:
            i += pd.Timedelta(dayofweek - i.dayofweek, "d")
            changed.append(i)
    return np.array(changed)


def plot_pdp_with_distribution(
    test_data,
    grid,
    ys,
    counts,
    reader,
    feature_name,
    individual,
    top_n_classes,
    datetime_level,
):
    feature_role = reader._roles[feature_name].name
    # I. Plot pdp
    sns.set(style="whitegrid", font_scale=1.5)
    fig, axs = plt.subplots(2, 1, figsize=(16, 12), gridspec_kw={"height_ratios": [3, 1]})
    axs[0].set_title("PDP: " + feature_name)
    n_classes = ys[0].shape[1]
    if n_classes == 1:
        data = pd.concat(
            [
                pd.DataFrame(
                    {
                        "x": grid[i],
                        "y": ys[i].ravel(),
                        "unit": np.arange(ys[i].shape[0]),
                    }
                )
                for i, _ in enumerate(grid)
            ]
        ).reset_index(drop=True)
        if feature_role in ["Numeric", "Datetime"]:
            if individual:
                g0 = sns.lineplot(
                    data=data[data["unit"] > 0],
                    x="x",
                    y="y",
                    ax=axs[0],
                    units="unit",
                    estimator=None,
                    alpha=0.05,
                    color="b",
                )
                g0 = sns.lineplot(
                    data=data[data["unit"] == 0],
                    x="x",
                    y="y",
                    ax=axs[0],
                    alpha=0.05,
                    color="b",
                    label="individual PDP",
                )
                g0 = sns.lineplot(
                    x=grid,
                    y=data.groupby("x").mean()["y"].values,
                    ax=axs[0],
                    linewidth=2,
                    color="r",
                    label="mean PDP",
                )
            else:
                g0 = sns.lineplot(data=data, x="x", y="y", ax=axs[0], color="b")
        else:
            g0 = sns.boxplot(data=data, x="x", y="y", ax=axs[0], showfliers=False, color="b")
    else:
        if reader.class_mapping:
            classes = sorted(reader.class_mapping, key=reader.class_mapping.get)[:top_n_classes]
        else:
            classes = np.arange(min(n_classes, top_n_classes))
        data = pd.concat(
            [
                pd.DataFrame({"x": grid[i], "y": ys[i][:, k], "class": name})
                for i, _ in enumerate(grid)
                for k, name in enumerate(classes)
            ]
        ).reset_index(drop=True)
        if reader._roles[feature_name].name in ["Numeric", "Datetime"]:
            g0 = sns.lineplot(data=data, x="x", y="y", hue="class", ax=axs[0])
        else:
            g0 = sns.boxplot(data=data, x="x", y="y", hue="class", ax=axs[0], showfliers=False)
    g0.set(ylabel="y_pred")
    # II. Plot distribution
    counts = np.array(counts) / sum(counts)
    if feature_role == "Numeric":
        g0.set(xlabel="feature value")
        g1 = sns.histplot(test_data[feature_name], kde=True, color="gray", ax=axs[1])
    elif feature_role == "Category":
        g0.set(xlabel=None)
        axs[0].set_xticklabels(grid, rotation=90)
        g1 = sns.barplot(x=grid, y=counts, ax=axs[1], color="gray")
    else:
        g0.set(xlabel=datetime_level)
        g1 = sns.barplot(x=grid, y=counts, ax=axs[1], color="gray")
    g1.set(xlabel=None)
    g1.set(ylabel="Frequency")
    g1.set(xticklabels=[])
    plt.tight_layout()
    plt.show()



"""Base classes to implement hyperparameter tuning."""

from abc import ABC
from abc import abstractmethod
from typing import Dict
from typing import Optional
from typing import Tuple
from typing import overload

from ...dataset.base import LAMLDataset

# if TYPE_CHECKING:
from ...ml_algo.base import MLAlgo
from ...validation.base import TrainValidIterator


class DistributionBase(ABC):
    """_summary_.

    Args:
        ABC (_type_): _description_
    """

    pass


class Choice(DistributionBase):
    """_summary_.

    Args:
        DistributionBase (_type_): _description_
    """

    def __init__(self, options) -> None:
        self.options = options


class Uniform(DistributionBase):
    """_summary_.

    Args:
        DistributionBase (_type_): _description_
    """

    def __init__(self, low, high, q=None, log=False) -> None:
        self.low = low
        self.high = high
        self.q = q
        self.log = log


class Normal(DistributionBase):
    """_summary_.

    Args:
        DistributionBase (_type_): _description_
    """

    def __init__(self, low, high, q=None, log=False) -> None:
        self.low = low
        self.high = high
        self.q = q
        self.log = log


class ParamsTuner(ABC):
    """Base abstract class for hyperparameters tuners."""

    _name: str = "AbstractTuner"
    _best_params: Dict = None
    _fit_on_holdout: bool = False  # if tuner should be fitted on holdout set

    @property
    def best_params(self) -> dict:
        """Get best params.

        Returns:
            Dict with best fitted params.

        """
        assert hasattr(self, "_best_params"), "ParamsTuner should be fitted first"
        return self._best_params

    @overload
    def fit(
        self,
        ml_algo: "MLAlgo",
        train_valid_iterator: Optional[TrainValidIterator] = None,
    ) -> Tuple["MLAlgo", LAMLDataset]:
        ...

    @abstractmethod
    def fit(
        self,
        ml_algo: "MLAlgo",
        train_valid_iterator: Optional[TrainValidIterator] = None,
    ) -> Tuple[None, None]:
        """Tune model hyperparameters.

        Args:
            ml_algo: ML algorithm.
            train_valid_iterator: Classic cv-iterator.

        Returns:
            (None, None) if ml_algo is fitted or models are not fitted during training,
            (BestMLAlgo, BestPredictionsLAMLDataset) otherwise.

        """


class DefaultTuner(ParamsTuner):
    """Default realization of ParamsTuner - just take algo's defaults."""

    _name: str = "DefaultTuner"

    def fit(
        self,
        ml_algo: "MLAlgo",
        train_valid_iterator: Optional[TrainValidIterator] = None,
    ) -> Tuple[None, None]:
        """Default fit method - just save defaults.

        Args:
            ml_algo: Algorithm that is tuned.
            train_valid_iterator: Empty.

        Returns:
            Tuple (None, None).
        """
        self._best_params = ml_algo.init_params_on_input(train_valid_iterator=train_valid_iterator)
        return None, None



"""Pipeline for tree based models."""

from typing import List
from typing import Optional
from typing import Union

import numpy as np

from ...dataset.np_pd_dataset import NumpyDataset
from ...dataset.np_pd_dataset import PandasDataset
from ...dataset.roles import CategoryRole
from ...dataset.roles import NumericRole
from ...transformers.base import ChangeRoles
from ...transformers.base import ColumnsSelector
from ...transformers.base import ConvertDataset
from ...transformers.base import LAMLTransformer
from ...transformers.base import SequentialTransformer
from ...transformers.base import SetAttribute
from ...transformers.base import UnionTransformer
from ...transformers.categorical import LabelEncoder
from ...transformers.categorical import OrdinalEncoder
from ...transformers.datetime import TimeToNum
from ...transformers.numeric import FillInf
from ...transformers.numeric import FillnaMedian
from ...transformers.numeric import NaNFlags
from ...transformers.numeric import StandardScaler
from ...transformers.seq import GetSeqTransformer
from ...transformers.seq import SeqDiffTransformer
from ...transformers.seq import SeqLagTransformer
from ...transformers.seq import SeqNumCountsTransformer
from ...transformers.seq import SeqStatisticsTransformer
from ..selection.base import ImportanceEstimator
from ..selection.base import SelectionPipeline
from ..utils import get_columns_by_role
from .base import FeaturesPipeline
from .base import TabularDataFeatures


NumpyOrPandas = Union[PandasDataset, NumpyDataset]


class LGBSimpleFeatures(FeaturesPipeline):
    """Creates simple pipeline for tree based models.

    Simple but is ok for select features.
    Numeric stay as is, Datetime transforms to numeric.
    Categorical label encoding.
    Maps input to output features exactly one-to-one.

    """

    def create_pipeline(self, train: NumpyOrPandas) -> LAMLTransformer:
        """Create tree pipeline.

        Args:
            train: Dataset with train features.

        Returns:
            Composite datetime, categorical, numeric transformer.

        """
        # TODO: Transformer params to config
        transformers_list = []

        # process categories
        categories = get_columns_by_role(train, "Category")
        if len(categories) > 0:
            cat_processing = SequentialTransformer(
                [
                    ColumnsSelector(keys=categories),
                    OrdinalEncoder(subs=None, random_state=42),
                    # ChangeRoles(NumericRole(np.float32))
                ]
            )
            transformers_list.append(cat_processing)

        # process datetimes
        datetimes = get_columns_by_role(train, "Datetime")
        if len(datetimes) > 0:
            dt_processing = SequentialTransformer([ColumnsSelector(keys=datetimes), TimeToNum()])
            transformers_list.append(dt_processing)

        # process numbers
        numerics = get_columns_by_role(train, "Numeric")
        if len(numerics) > 0:
            num_processing = SequentialTransformer(
                [
                    ColumnsSelector(keys=numerics),
                    ConvertDataset(dataset_type=NumpyDataset),
                ]
            )
            transformers_list.append(num_processing)

        union_all = UnionTransformer(transformers_list)

        return union_all


class LGBSeqSimpleFeatures(FeaturesPipeline, TabularDataFeatures):
    """LGBSeqSimpleFeatures.

    Args:
        feats_imp: Features importances mapping.
        top_intersections: Max number of categories
          to generate intersections.
        max_intersection_depth: Max depth of cat intersection.
        subsample: Subsample to calc data statistics.
        multiclass_te_co: Cutoff if use target encoding in cat
          handling on multiclass task if number of classes is high.
        auto_unique_co: Switch to target encoding if high cardinality.

    """

    def __init__(
        self,
        feats_imp: Optional[ImportanceEstimator] = None,
        top_intersections: int = 5,
        max_intersection_depth: int = 3,
        subsample: Optional[Union[int, float]] = None,
        multiclass_te_co: int = 3,
        auto_unique_co: int = 10,
        output_categories: bool = False,
        fill_na=False,
        scaler=False,
        transformers_params=None,
        **kwargs
    ):
        super().__init__(
            multiclass_te_co=multiclass_te_co,
            top_intersections=top_intersections,
            max_intersection_depth=max_intersection_depth,
            subsample=subsample,
            feats_imp=feats_imp,
            auto_unique_co=auto_unique_co,
            output_categories=output_categories,
            ascending_by_cardinality=False,
        )

        self.fill_na = fill_na
        self.scaler = scaler
        self.transformers_params = transformers_params

    def get_seq_pipeline(self, train):
        """Create pipeline for seq data.

        Args:
            train: Dataset with train features.

        Returns:
            Composite datetime, categorical, numeric transformer.

        """
        # process datetimes
        time_transformers_list = []

        # datetimes features generating
        time_features_transformers_list = []
        datetimes = get_columns_by_role(train, "Datetime")
        if len(datetimes) > 0:
            dt_processing = SequentialTransformer([ColumnsSelector(keys=datetimes), TimeToNum()])
            time_features_transformers_list.append(dt_processing)
            time_features_transformers_list.append(self.get_datetime_diffs(train))
            time_features_transformers_list.append(self.get_datetime_seasons(train, NumericRole(np.float32)))

        # datetime features preprocessing
        time_preprocessing_transformers_list = []
        if self.fill_na:
            time_preprocessing_transformers_list.append(
                UnionTransformer([SequentialTransformer([FillInf(), FillnaMedian()]), NaNFlags()])
            )

            if self.scaler:
                time_preprocessing_transformers_list.append(StandardScaler())

        time_transformers_list.append(UnionTransformer(time_features_transformers_list))
        time_transformers_list += time_preprocessing_transformers_list

        time_transforms = SequentialTransformer(time_transformers_list)

        if self.transformers_params["lag_time_features"]:
            seq = ColumnsSelector(keys=[])  # SequentialTransformer([EmptyTransformer(), ColumnsSelector(keys=[])])
            time_transforms = SequentialTransformer(
                [
                    UnionTransformer([time_transforms, seq]),
                    SeqLagTransformer(lags=self.transformers_params["lag_time_features"]),
                ]
            )

        # process other features
        other_transformers_list = []

        # process categories
        other_features_transformers_list = []
        categories = get_columns_by_role(train, "Category")
        if len(categories) > 0:
            cat_processing = SequentialTransformer(
                [
                    ColumnsSelector(keys=categories),
                    LabelEncoder(subs=None, random_state=42),
                    ChangeRoles(NumericRole(np.float32)),
                ]
            )
            other_features_transformers_list.append(cat_processing)

        # process numeric
        numerics = get_columns_by_role(train, "Numeric")
        if len(numerics) > 0:
            num_processing = SequentialTransformer(
                [ColumnsSelector(keys=numerics), ConvertDataset(dataset_type=NumpyDataset)]
            )
            other_features_transformers_list.append(num_processing)

        # other features preprocessing
        other_preprocessing_transformers_list = []
        if self.fill_na:
            other_preprocessing_transformers_list.append(
                UnionTransformer([SequentialTransformer([FillInf(), FillnaMedian()]), NaNFlags()])
            )

            if self.scaler:
                other_preprocessing_transformers_list.append(StandardScaler())

        other_transformers_list.append(UnionTransformer(other_features_transformers_list))
        other_transformers_list += other_preprocessing_transformers_list

        lags = self.transformers_params["lag_features"]
        diffs = self.transformers_params["diff_features"]

        other_transforms = SequentialTransformer(other_transformers_list)

        if lags or diffs:
            seq = ColumnsSelector(keys=[])  # SequentialTransformer([EmptyTransformer(), ColumnsSelector(keys=[])])
            seq_features = []

            if lags:
                seq_features.append(SeqLagTransformer(lags=lags))

            if diffs:
                # if we have lag with number 0, we shouldn't have diff with number 0
                if lags:
                    flag_del_0_diff = not (
                        not isinstance(diffs, int) and 0 not in diffs or not isinstance(lags, int) and 0 not in lags
                    )
                else:
                    flag_del_0_diff = False
                seq_features.append(SeqDiffTransformer(diffs=diffs, flag_del_0_diff=flag_del_0_diff))

            other_transforms = SequentialTransformer(
                [UnionTransformer([other_transforms, seq]), UnionTransformer(seq_features)]
            )

        all_feats = SequentialTransformer(
            [
                GetSeqTransformer(name=train.name),
                SetAttribute("date", datetimes[0]),
                UnionTransformer([time_transforms, other_transforms]),
            ]
        )

        return all_feats

    def create_pipeline(self, train: NumpyOrPandas) -> LAMLTransformer:
        """Create tree pipeline.

        Args:
            train: Dataset with train features.

        Returns:
            Composite datetime, categorical, numeric transformer.

        """
        # TODO: Transformer params to config

        transformers_list = []

        # process categories
        categories = get_columns_by_role(train, "Category")
        if len(categories) > 0:
            cat_processing = SequentialTransformer(
                [
                    ColumnsSelector(keys=categories),
                    OrdinalEncoder(subs=None, random_state=42),
                    # ChangeRoles(NumericRole(np.float32))
                ]
            )
            transformers_list.append(cat_processing)

        # process datetimes
        datetimes = get_columns_by_role(train, "Datetime")
        if len(datetimes) > 0:
            dt_processing = SequentialTransformer([ColumnsSelector(keys=datetimes), TimeToNum()])
            transformers_list.append(dt_processing)

        # process numbers
        numerics = get_columns_by_role(train, "Numeric")
        if len(numerics) > 0:
            num_processing = SequentialTransformer(
                [ColumnsSelector(keys=numerics), ConvertDataset(dataset_type=NumpyDataset)]
            )
            transformers_list.append(num_processing)

        union_all = UnionTransformer(transformers_list)

        if hasattr(train, "seq_data"):
            if train.seq_data is not None:
                seq_pipes = []
                for name, seq_data in train.seq_data.items():
                    seq_pipes.append(self.get_seq_pipeline(seq_data))
                union_all = UnionTransformer(seq_pipes + [union_all])

        # dummy to get task metadata
        union_all = UnionTransformer([ColumnsSelector(keys=[]), union_all])
        union_all = SequentialTransformer([union_all, ConvertDataset(dataset_type=NumpyDataset)])

        return union_all


class LGBMultiSeqSimpleFeatures(FeaturesPipeline, TabularDataFeatures):
    """LGBMultiSeqSimpleFeatures.

    Args:
        feats_imp: Features importances mapping.
        top_intersections: Max number of categories
          to generate intersections.
        max_intersection_depth: Max depth of cat intersection.
        subsample: Subsample to calc data statistics.
        multiclass_te_co: Cutoff if use target encoding in cat
          handling on multiclass task if number of classes is high.
        auto_unique_co: Switch to target encoding if high cardinality.

    """

    def __init__(
        self,
        feats_imp: Optional[ImportanceEstimator] = None,
        top_intersections: int = 5,
        max_intersection_depth: int = 3,
        subsample: Optional[Union[int, float]] = None,
        multiclass_te_co: int = 3,
        auto_unique_co: int = 10,
        output_categories: bool = False,
        **kwargs
    ):

        super().__init__(
            multiclass_te_co=multiclass_te_co,
            top_intersections=top_intersections,
            max_intersection_depth=max_intersection_depth,
            subsample=subsample,
            feats_imp=feats_imp,
            auto_unique_co=auto_unique_co,
            output_categories=output_categories,
            ascending_by_cardinality=False,
        )

    def get_seq_pipeline(self, train):
        """Create pipeline for seq data.

        Args:
            train: Dataset with train features.

        Returns:
            Composite datetime, categorical, numeric transformer.

        """
        transformers_list = []
        # process categories
        categories = get_columns_by_role(train, "Category")
        if len(categories) > 0:
            cat_processing = SequentialTransformer(
                [
                    ColumnsSelector(keys=categories),
                    LabelEncoder(subs=None, random_state=42),
                    ChangeRoles(NumericRole(np.float32)),
                ]
            )
            transformers_list.append(cat_processing)

        # process datetimes
        datetimes = get_columns_by_role(train, "Datetime")
        if len(datetimes) > 0:
            dt_processing = SequentialTransformer([ColumnsSelector(keys=datetimes), TimeToNum()])
            transformers_list.append(dt_processing)
            transformers_list.append(self.get_datetime_diffs(train))
            transformers_list.append(self.get_datetime_seasons(train, NumericRole(np.float32)))

        numerics = get_columns_by_role(train, "Numeric")
        if len(numerics) > 0:
            num_processing = SequentialTransformer(
                [ColumnsSelector(keys=numerics), ConvertDataset(dataset_type=NumpyDataset)]
            )
            transformers_list.append(num_processing)

        simple_seq_transforms = UnionTransformer(transformers_list)

        # to seq dataset
        simple_seq_transforms = UnionTransformer([ColumnsSelector(keys=[]), simple_seq_transforms])

        # get seq features
        if train.scheme is not None:
            if train.scheme.get("type", "full") == "lookup":
                seq = SeqStatisticsTransformer()
            else:
                seq = SeqNumCountsTransformer()
        else:
            seq = SeqNumCountsTransformer()

        all_feats = SequentialTransformer(
            [GetSeqTransformer(name=train.name), simple_seq_transforms, seq]  # preprocessing  # plain features
        )

        return all_feats

    def create_pipeline(self, train: NumpyOrPandas) -> LAMLTransformer:
        """Create tree pipeline.

        Args:
            train: Dataset with train features.

        Returns:
            Composite datetime, categorical, numeric transformer.

        """
        # TODO: Transformer params to config

        transformers_list = []

        # process categories
        categories = get_columns_by_role(train, "Category")
        if len(categories) > 0:
            cat_processing = SequentialTransformer(
                [
                    ColumnsSelector(keys=categories),
                    OrdinalEncoder(subs=None, random_state=42),
                    # ChangeRoles(NumericRole(np.float32))
                ]
            )
            transformers_list.append(cat_processing)

        # process datetimes
        datetimes = get_columns_by_role(train, "Datetime")
        if len(datetimes) > 0:
            dt_processing = SequentialTransformer([ColumnsSelector(keys=datetimes), TimeToNum()])
            transformers_list.append(dt_processing)

        # process numbers
        numerics = get_columns_by_role(train, "Numeric")
        if len(numerics) > 0:
            num_processing = SequentialTransformer(
                [ColumnsSelector(keys=numerics), ConvertDataset(dataset_type=NumpyDataset)]
            )
            transformers_list.append(num_processing)

        union_all = UnionTransformer(transformers_list)

        if hasattr(train, "seq_data"):
            if train.seq_data is not None:
                seq_pipes = []
                for name, seq_data in train.seq_data.items():
                    seq_pipes.append(self.get_seq_pipeline(seq_data))
                union_all = UnionTransformer(seq_pipes + [union_all])

        # dummy to get task metadata
        union_all = UnionTransformer([ColumnsSelector(keys=[]), union_all])

        return union_all


class LGBAdvancedPipeline(FeaturesPipeline, TabularDataFeatures):
    """Create advanced pipeline for trees based models.

    Includes:

        - Different cats and numbers handling according to role params.
        - Dates handling - extracting seasons and create datediffs.
        - Create categorical intersections.


    Args:
        feats_imp: Features importances mapping.
        top_intersections: Max number of categories
            to generate intersections.
        max_intersection_depth: Max depth of cat intersection.
        subsample: Subsample to calc data statistics.
        multiclass_te_co: Cutoff if use target encoding in cat
            handling on multiclass task if number of classes is high.
        auto_unique_co: Switch to target encoding if high cardinality.

    """

    def __init__(
        self,
        feats_imp: Optional[Union[ImportanceEstimator, SelectionPipeline]] = None,
        top_intersections: int = 5,
        max_intersection_depth: int = 3,
        subsample: Optional[Union[int, float]] = None,
        multiclass_te_co: int = 3,
        auto_unique_co: int = 10,
        output_categories: bool = False,
        fill_na: bool = False,
        ascending_by_cardinality: bool = False,
        use_groupby: bool = False,
        groupby_types: List[str] = ["delta_median", "delta_mean", "min", "max", "std", "mode", "is_mode"],
        groupby_triplets: list = [],
        groupby_top_based_on: str = "cardinality",
        groupby_top_categorical: int = 3,
        groupby_top_numerical: int = 3,
        **kwargs
    ):
        super().__init__(
            multiclass_te_co=multiclass_te_co,
            top_intersections=top_intersections,
            max_intersection_depth=max_intersection_depth,
            subsample=subsample,
            feats_imp=feats_imp,
            auto_unique_co=auto_unique_co,
            output_categories=output_categories,
            ascending_by_cardinality=ascending_by_cardinality,
            groupby_types=groupby_types,
            groupby_triplets=groupby_triplets,
            groupby_top_based_on=groupby_top_based_on,
            groupby_top_categorical=groupby_top_categorical,
            groupby_top_numerical=groupby_top_numerical,
        )
        self.fill_na = fill_na
        self.use_groupby = use_groupby

    def create_pipeline(self, train: NumpyOrPandas) -> LAMLTransformer:
        """Create tree pipeline.

        Args:
            train: Dataset with train features.

        Returns:
            Transformer.

        """
        transformer_list = []
        target_encoder = self.get_target_encoder(train)

        output_category_role = (
            CategoryRole(np.float32, label_encoded=True) if self.output_categories else NumericRole(np.float32)
        )

        # handle categorical feats
        # split categories by handling type. This pipe use 3 encodings - freq/label/target/ordinal
        # 1 - separate freqs. It does not need label encoding
        transformer_list.append(self.get_freq_encoding(train))

        # 2 - check different target encoding parts and split (ohe is the same as auto - no ohe in gbm)
        auto = get_columns_by_role(train, "Category", encoding_type="auto") + get_columns_by_role(
            train, "Category", encoding_type="ohe"
        )

        if self.output_categories:
            le = (
                auto
                + get_columns_by_role(train, "Category", encoding_type="oof")
                + get_columns_by_role(train, "Category", encoding_type="int")
            )
            te_lst = []
            ordinal = None

        else:
            le = get_columns_by_role(train, "Category", encoding_type="int")
            ordinal = get_columns_by_role(train, "Category", ordinal=True)

            if target_encoder is not None:
                te_lst = get_columns_by_role(train, "Category", encoding_type="oof")
                # split auto categories by unique values cnt
                un_values = self.get_uniques_cnt(train, auto)
                te_lst = te_lst + [x for x in un_values.index if un_values[x] > self.auto_unique_co]
                ordinal = ordinal + list(set(auto) - set(te_lst))

            else:
                te_lst = []
                ordinal = ordinal + auto + get_columns_by_role(train, "Category", encoding_type="oof")

            ordinal = sorted(list(set(ordinal)))

        # get label encoded categories
        le_part = self.get_categorical_raw(train, le)
        if le_part is not None:
            le_part = SequentialTransformer([le_part, ChangeRoles(output_category_role)])
            transformer_list.append(le_part)

        # get target encoded part
        te_part = self.get_categorical_raw(train, te_lst)
        if te_part is not None:
            te_part = SequentialTransformer([te_part, target_encoder()])
            transformer_list.append(te_part)

        # get intersection of top categories
        intersections = self.get_categorical_intersections(train)
        if intersections is not None:
            if target_encoder is not None:
                ints_part = SequentialTransformer([intersections, target_encoder()])
            else:
                ints_part = SequentialTransformer([intersections, ChangeRoles(output_category_role)])

            transformer_list.append(ints_part)

        # add numeric pipeline
        transformer_list.append(self.get_numeric_data(train))
        transformer_list.append(self.get_ordinal_encoding(train, ordinal))
        # add difference with base date
        transformer_list.append(self.get_datetime_diffs(train))
        # add datetime seasonality
        transformer_list.append(self.get_datetime_seasons(train, NumericRole(np.float32)))

        if self.use_groupby:
            transformer_list.append(self.get_groupby(train))

        # final pipeline
        union_all = UnionTransformer([x for x in transformer_list if x is not None])
        if self.fill_na:
            union_all = SequentialTransformer([union_all, SequentialTransformer([FillInf(), FillnaMedian()])])

        return union_all



"""Whitebox MLPipeline."""

import warnings

from typing import Tuple
from typing import Union
from typing import cast

from ...dataset.np_pd_dataset import NumpyDataset
from ...dataset.np_pd_dataset import PandasDataset
from ...ml_algo.tuning.base import ParamsTuner
from ...ml_algo.whitebox import WbMLAlgo
from ...validation.base import DummyIterator
from ...validation.base import TrainValidIterator
from ..features.wb_pipeline import WBFeatures
from ..selection.base import EmptySelector
from .base import MLPipeline


TunedWB = Union[WbMLAlgo, Tuple[WbMLAlgo, ParamsTuner]]


class WBPipeline(MLPipeline):
    """Special pipeline to handle WhiteBox model."""

    @property
    def whitebox(self) -> WbMLAlgo:
        """Return first whitebox model."""
        if len(self.ml_algos[0].models) > 1:
            warnings.warn("More than 1 whitebox model is fitted during cross validation. Only first is returned")

        return self.ml_algos[0].models[0]

    def __init__(self, whitebox: TunedWB):
        """Create WhiteBox MLPipeline.

        Args:
            whitebox: WhiteBox model.

        """
        super().__init__([whitebox], True, features_pipeline=WBFeatures())
        self._used_features = None

    def fit_predict(self, train_valid: TrainValidIterator) -> NumpyDataset:
        """Fit WhiteBox.

        Args:
            train_valid: Classic cv-iterator.

        Returns:
            Dataset.

        """
        _subsamp_to_refit = train_valid.train[:5]
        val_pred = super().fit_predict(train_valid)
        self._prune_pipelines(_subsamp_to_refit)

        return cast(NumpyDataset, val_pred)

    def predict(self, dataset: PandasDataset, report: bool = False) -> NumpyDataset:
        """Predict WhiteBox.

        Additional report param stands for WhiteBox report generation.

        Args:
            dataset: Dataset of text features.
            report: Flag if generate report.

        Returns:
            Dataset.

        """
        dataset = self.features_pipeline.transform(dataset)
        args = []
        if self.ml_algos[0].params["report"]:
            args = [report]
        pred = self.ml_algos[0].predict(dataset, *args)

        return pred

    def _prune_pipelines(self, subsamp: PandasDataset):
        # upd used features attribute from list of whiteboxes
        feats_from_wb = set.union(*[set(list(x.features_fit.index)) for x in self.ml_algos[0].models])
        # cols wo prefix - numerics and categories
        raw_columns = list(set(subsamp.features).intersection(feats_from_wb))
        diff_cols = list(set(feats_from_wb).difference(subsamp.features))

        seasons = ["__".join(x.split("__")[1:]) for x in diff_cols if x.startswith("season_")]

        base_diff = [x.split("__") for x in diff_cols if x.startswith("basediff_")]
        base_diff = [("_".join(x[0].split("_")[1:]), "__".join(x[1:])) for x in base_diff]
        base_dates, compare_dates = [x[0] for x in base_diff], [x[1] for x in base_diff]
        dates = list(set(base_dates + compare_dates + seasons))

        raw_columns.extend(dates)

        subsamp = subsamp[:, raw_columns]
        self.features_pipeline = WBFeatures()

        self.pre_selection = EmptySelector()
        self.post_selection = EmptySelector()

        train_valid = DummyIterator(subsamp)
        train_valid = train_valid.apply_selector(self.pre_selection)
        train_valid = train_valid.apply_feature_pipeline(self.features_pipeline)
        train_valid.apply_selector(self.post_selection)



"""Nested MLPipeline."""

import logging

from copy import copy
from copy import deepcopy
from typing import Any
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

import numpy as np
import pandas as pd

from pandas import Series

from ...dataset.np_pd_dataset import NumpyDataset
from ...ml_algo.base import PandasDataset
from ...ml_algo.base import TabularDataset
from ...ml_algo.base import TabularMLAlgo
from ...ml_algo.tuning.base import DefaultTuner
from ...ml_algo.tuning.base import ParamsTuner
from ...ml_algo.utils import tune_and_fit_predict
from ...reader.utils import set_sklearn_folds
from ...utils.timer import PipelineTimer
from ...validation.base import TrainValidIterator
from ...validation.utils import create_validation_iterator
from ..features.base import FeaturesPipeline
from ..selection.base import SelectionPipeline
from ..selection.importance_based import ImportanceEstimator
from .base import MLPipeline


logger = logging.getLogger(__name__)


class NestedTabularMLAlgo(TabularMLAlgo, ImportanceEstimator):
    """Wrapper for MLAlgo to make it trainable over nested folds.

    Limitations - only for ``TabularMLAlgo``.

    """

    @property
    def params(self) -> dict:
        """Parameters of ml_algo."""
        if self._ml_algo._params is None:
            self._ml_algo._params = copy(self.default_params)
        return self._ml_algo._params

    @params.setter
    def params(self, new_params: dict):
        assert isinstance(new_params, dict)
        self._ml_algo.params = {**self._ml_algo.params, **new_params}
        self._params = self._ml_algo.params

    def init_params_on_input(self, train_valid_iterator: TrainValidIterator) -> dict:
        """Init params depending on input data.

        Args:
            train_valid_iterator: Iterator over input data.

        Returns:
            dict with model hyperparameters.

        """
        return self._ml_algo.init_params_on_input(train_valid_iterator)

    def __init__(
        self,
        ml_algo: TabularMLAlgo,
        tuner: Optional[ParamsTuner] = None,
        refit_tuner: bool = False,
        cv: int = 5,
        n_folds: Optional[int] = None,
    ):
        self._name = ml_algo.name
        self._default_params = ml_algo.default_params

        super().__init__(default_params=ml_algo.default_params)
        self.default_params = ml_algo.default_params

        self._params_tuner = tuner
        self._refit_tuner = refit_tuner

        # take timer from inner algo and set to outer
        self.timer = ml_algo.timer
        if self.timer.key is not None:
            self.timer.key = "nested_" + self.timer.key
        # reset inner timer
        self._ml_algo = ml_algo.set_timer(PipelineTimer().start().get_task_timer())

        self.nested_cv = cv
        self.n_folds = n_folds

    def fit_predict(self, train_valid_iterator: TrainValidIterator) -> NumpyDataset:  # noqa DAR102
        self.timer.start()
        div = len(train_valid_iterator) if self.n_folds is None else self.n_folds
        self._per_task_timer = self.timer.time_left / div

        return super().fit_predict(train_valid_iterator)

    def fit_predict_single_fold(self, train: TabularDataset, valid: TabularDataset) -> Tuple[Any, np.ndarray]:
        """Implements training and prediction on single fold.

        Args:
            train: TabularDataset to train.
            valid: TabularDataset to validate.

        Returns:
            Tuple (model, predicted_values).

        """
        logger.info3("Start fit_predict for nested model on a single fold ...")
        # TODO: rewrite
        if isinstance(train, PandasDataset):
            train.folds = pd.Series(
                set_sklearn_folds(
                    train.task,
                    train.target,
                    self.nested_cv,
                    random_state=42,
                    group=train.group,
                )
            )
        else:
            train.folds = set_sklearn_folds(
                train.task,
                train.target,
                self.nested_cv,
                random_state=42,
                group=train.group,
            )

        train_valid = create_validation_iterator(train, n_folds=self.n_folds)

        model = deepcopy(self._ml_algo)
        model.set_timer(PipelineTimer(timeout=self._per_task_timer, overhead=0).start().get_task_timer())
        logger.debug(self._ml_algo.params)
        tuner = self._params_tuner
        if self._refit_tuner:
            tuner = deepcopy(tuner)

        if tuner is None:
            logger.debug("Run without tuner")
            model.fit_predict(train_valid)
        else:
            logger.debug("Run with tuner")
            (
                model,
                _,
            ) = tune_and_fit_predict(model, tuner, train_valid, True)

        val_pred = model.predict(valid).data
        logger.debug("Model params", model.params)
        return model, val_pred

    def predict_single_fold(self, model: Any, dataset: TabularDataset) -> np.ndarray:
        """Model prediction on a dataset.

        Args:
            model: Model.
            dataset: Dataset.

        Returns:
            Predictions.

        """
        pred = model.predict(dataset).data

        return pred

    def _get_search_spaces(self, suggested_params: dict, estimated_n_trials: int) -> dict:
        return self._ml_algo._get_search_spaces(suggested_params, estimated_n_trials)

    def get_features_score(self) -> Series:
        """Score of each features."""
        scores = pd.concat([x.get_features_score() for x in self.models], axis=1).mean(axis=1)

        return scores

    def fit(self, train_valid: TrainValidIterator):
        """Just to be compatible with :class:`~lightautoml.validation.selection.importance_based.ImportanceEstimator`.

        Args:
            train_valid: Classic cv iterator.

        """
        self.fit_predict(train_valid)


class NestedTabularMLPipeline(MLPipeline):
    """Wrapper for MLPipeline to make it trainable over nested folds.

    Limitations:

        - Only for TabularMLAlgo
        - Nested trained only MLAlgo. FeaturesPipelines and
          SelectionPipelines are trained as usual.

    Args:
        ml_algos: Sequence of MLAlgo's or Pair - (MlAlgo, ParamsTuner).
        force_calc: Flag if single fold of
            :class:`~lightautoml.ml_algo.base.MlAlgo`
            should be calculated anyway.
        pre_selection: Initial feature selection.
            If ``None`` there is no initial selection.
        features_pipeline: Composition of feature transforms.
        post_selection: Post feature selection.
            If ``None`` there is no post selection.
        cv: Nested folds cv split.
        n_folds: Limit of valid iterations from cv.
        inner_tune: Should we refit tuner each inner
            cv run or tune ones on outer cv.
        refit_tuner: Should we refit tuner each inner
            loop with ``inner_tune==True``.

    """

    def __init__(
        self,
        ml_algos: Sequence[Union[TabularMLAlgo, Tuple[TabularMLAlgo, ParamsTuner]]],
        force_calc: Union[bool, Sequence[bool]] = True,
        pre_selection: Optional[SelectionPipeline] = None,
        features_pipeline: Optional[FeaturesPipeline] = None,
        post_selection: Optional[SelectionPipeline] = None,
        cv: int = 1,
        n_folds: Optional[int] = None,
        inner_tune: bool = False,
        refit_tuner: bool = False,
    ):
        if cv > 1:
            new_ml_algos = []

            for n, mt_pair in enumerate(ml_algos):

                try:
                    mod, tuner = mt_pair
                except (TypeError, ValueError):
                    mod, tuner = mt_pair, DefaultTuner()

                if inner_tune:
                    new_ml_algos.append(NestedTabularMLAlgo(mod, tuner, refit_tuner, cv, n_folds))
                else:
                    new_ml_algos.append((NestedTabularMLAlgo(mod, None, True, cv, n_folds), tuner))

            ml_algos = new_ml_algos

        super().__init__(ml_algos, force_calc, pre_selection, features_pipeline, post_selection)



from typing import Optional

import torch
import torch.nn.functional as F

from torch import nn
from torch.distributions.utils import clamp_probs

from .data_process import create_emb_layer


class TIModel(nn.Module):
    def __init__(
        self,
        voc_size: int,
        embed_dim: int = 50,
        conv_filters: int = 100,
        conv_ksize: int = 3,
        drop_rate: float = 0.2,
        hidden_dim: int = 100,
        weights_matrix: Optional[torch.FloatTensor] = None,
        trainable_embeds: bool = False,
    ):
        super(TIModel, self).__init__()

        self.lookup = create_emb_layer(weights_matrix, voc_size, embed_dim, trainable_embeds)

        embed_dim = self.lookup.embedding_dim
        self.drop1 = nn.Dropout(p=drop_rate)
        self.conv1 = nn.Conv1d(embed_dim, conv_filters, conv_ksize, padding=1)
        self.act1 = nn.ReLU()
        self.pool1 = nn.AdaptiveAvgPool1d(output_size=1)
        self.global_info = nn.Linear(conv_filters, hidden_dim)
        self.global_act = nn.ReLU()
        self.conv2 = nn.Conv1d(conv_filters, hidden_dim, conv_ksize, padding=1)
        self.act2 = nn.ReLU()
        self.local_info = nn.Conv1d(hidden_dim, hidden_dim, conv_ksize, padding=1)
        self.local_act = nn.ReLU()
        self.drop3 = nn.Dropout(p=drop_rate)
        self.conv3 = nn.Conv1d(2 * hidden_dim, conv_filters, 1)
        self.act3 = nn.ReLU()
        self.conv4 = nn.Conv1d(conv_filters, 1, 1)

    def get_embedding(self, input):
        return self.lookup(input)

    def get_embedding_layer(self):
        return self.lookup

    def freeze_embedding(self):
        embeds = self.get_embedding_layer()
        embeds.weight.requires_grad_(False)

    def unfreeze_embedding(self):
        embeds = self.get_embedding_layer()
        embeds.weight.requires_grad_(True)

    def predict(self, x):
        # input: batch_size x len_seq x embed_dim
        # batch_size x embed_dim x len_seq
        x = x.transpose(1, 2)
        # batch_size x conv_filters x len_seq
        x = self.act1(self.conv1(self.drop1(x)))
        # batch_size x conv_filters x len_seq -> batch_size x conv_filters ->
        # -> batch_size x hidden_dim
        global_info = self.global_act(self.global_info(self.pool1(x).squeeze(2)))
        # batch_size x hidden_dim x len_seq ->
        # -> batch_size x hidden_dim x len_seq
        local_info = self.local_act(self.local_info(self.act2(self.conv2(x))))
        # batch_size x hidden_dim x 1 ->
        # -> batch_size x hidden_dim x len_seq
        global_info = global_info.unsqueeze(-1).expand_as(local_info)
        # batch_size x 2 * hidden_dim x len_seq
        z = torch.cat([global_info, local_info], dim=1)
        # batch_size x conv_filters x len_seq
        z = self.act3(self.conv3(self.drop3(z)))
        # batch_size x 1 x len_seq
        logits = self.conv4(z)

        return logits

    def forward(self, x):
        embed = self.get_embedding(x)
        logits = self.predict(embed)

        return logits


class GumbelTopKSampler(nn.Module):
    def __init__(self, T, k):
        super(GumbelTopKSampler, self).__init__()
        # Use as parameter to serialize,
        # Serialization of pytorch, when loading an already initialized model
        # does not load class arguments, only arguments that are model parameters.
        # So code (pseudo-code) like this, doesn't work right:
        # >>> l2x = L2XModel(...)
        # >>> for epoch in range(n_epochs):
        # >>>     train_step(...)
        # >>>     val_loss = valid_step(...)
        # >>>     if best_loss > val_loss:
        # >>>         torch.save(l2x.state_dict(), self._checkpoint_path)
        # >>> l2x.load_state_dict(torch.load(self._checkpoint_path))
        self.T = nn.Parameter(torch.tensor(T, dtype=torch.float32), requires_grad=False)
        self.k = nn.Parameter(torch.tensor(k, dtype=torch.int32), requires_grad=False)

    def sample_continous(self, logits):
        l_shape = (logits.shape[0], self.k, logits.shape[2])
        u = clamp_probs(torch.rand(l_shape, device=logits.device))
        gumbel = -torch.log(-torch.log(u))
        noisy_logits = (gumbel + logits) / self.T
        samples = F.softmax(noisy_logits, dim=-1)
        samples = torch.max(samples, dim=1)[0]

        return samples

    def sample_discrete(self, logits):
        threshold = torch.topk(logits, self.k, sorted=True)[0][..., -1]
        samples = torch.ge(logits.squeeze(1), threshold).float()

        return samples

    def forward(self, logits):
        csamples = None
        if self.training:
            csamples = self.sample_continous(logits)

        dsamples = self.sample_discrete(logits)

        return dsamples, csamples


class SoftSubSampler(nn.Module):
    def __init__(self, T, k):
        super(SoftSubSampler, self).__init__()
        self.T = nn.Parameter(torch.tensor(T, dtype=torch.float32), requires_grad=False)
        self.k = nn.Parameter(torch.tensor(k, dtype=torch.int32), requires_grad=False)

    def inject_noise(self, logits):
        u = clamp_probs(torch.rand_like(logits))
        z = -torch.log(-torch.log(u))
        noisy_logits = logits + z
        return noisy_logits

    def continuous_topk(self, w, separate=False):
        khot_list = []
        onehot_approx = torch.zeros_like(w, dtype=torch.float32)
        for _ in range(self.k):
            khot_mask = clamp_probs(1.0 - onehot_approx)
            w += torch.log(khot_mask)
            onehot_approx = F.softmax(w / self.T, dim=-1)
            khot_list.append(onehot_approx)
        if separate:
            return khot_list
        else:
            return torch.stack(khot_list, dim=-1).sum(-1).squeeze(1)

    def sample_continous(self, logits):
        return self.continuous_topk(self.inject_noise(logits))

    def sample_discrete(self, logits):
        threshold = torch.topk(logits, self.k, sorted=True)[0][..., -1]
        samples = torch.ge(logits.squeeze(1), threshold).float()

        return samples

    def forward(self, logits):
        csamples = None
        if self.training:
            csamples = self.sample_continous(logits)

        dsamples = self.sample_discrete(logits)

        return dsamples, csamples


class DistilPredictor(nn.Module):
    def __init__(
        self,
        task_name: str,
        n_outs: int,
        voc_size: int,
        embed_dim: int = 300,
        hidden_dim: int = 100,
        weights_matrix: Optional[torch.FloatTensor] = None,
        trainable_embeds: bool = False,
    ):
        super(DistilPredictor, self).__init__()

        self.lookup = create_emb_layer(weights_matrix, voc_size, embed_dim, trainable_embeds)
        embed_dim = self.lookup.embedding_dim
        self.fc1 = nn.Linear(embed_dim, hidden_dim)
        self.act = nn.ReLU()
        if task_name == "reg":
            self.head = nn.Linear(hidden_dim, n_outs)
        elif task_name == "binary":
            self.head = nn.Sequential(nn.Linear(hidden_dim, n_outs), nn.Sigmoid())
        elif task_name == "multiclass":
            self.head = nn.Sequential(nn.Linear(hidden_dim, n_outs), nn.Softmax(dim=-1))

    def get_embedding(self, input):
        return self.lookup(input)

    def get_embedding_layer(self):
        return self.lookup

    def freeze_embedding(self):
        embeds = self.get_embedding_layer()
        embeds.weight.requires_grad_(False)

    def unfreeze_embedding(self):
        embeds = self.get_embedding_layer()
        embeds.weight.requires_grad_(True)

    def predict(self, embed, T):
        out = torch.mean(embed * T.unsqueeze(2), axis=1)
        out = self.act(self.fc1(out))
        out = self.head(out)

        return out

    def forward(self, x, T):
        embed = self.get_embedding(x)
        out = self.predict(embed, T)

        return out


class L2XModel(nn.Module):
    def __init__(
        self,
        task_name: str,
        n_outs: int,
        voc_size: int = 1000,
        embed_dim: int = 100,
        conv_filters: int = 100,
        conv_ksize: int = 3,
        drop_rate: float = 0.2,
        hidden_dim: int = 100,
        T: float = 0.3,
        k: int = 5,
        weights_matrix: Optional[torch.FloatTensor] = None,
        trainable_embeds: bool = False,
        sampler: str = "gumbeltopk",
        anneal_factor: float = 1.0,
    ):
        super(L2XModel, self).__init__()

        self.ti_model = TIModel(
            voc_size,
            embed_dim,
            conv_filters,
            conv_ksize,
            drop_rate,
            hidden_dim,
            weights_matrix,
            trainable_embeds,
        )  # token importance model
        self.T = T
        self.anneal_factor = anneal_factor
        if sampler == "gumbeltopk":
            self.sampler = GumbelTopKSampler(T, k)
        else:
            self.sampler = SoftSubSampler(T, k)
        self.distil_model = DistilPredictor(
            task_name,
            n_outs,
            voc_size,
            embed_dim,
            hidden_dim,
            weights_matrix,
            trainable_embeds,
        )

    def forward(self, x):
        """Forward pass."""
        logits = self.ti_model(x)
        dsamples, csamples = self.sampler(logits)
        if self.training:
            T = csamples
        else:
            T = dsamples
        out = self.distil_model(x, T)

        return out, T

    def anneal(self):
        """Temperature annealing."""
        self.sampler.T *= self.anneal_factor



"""Role contains information about the column, which determines how it is processed."""

from datetime import datetime
from typing import Any
from typing import Callable
from typing import Optional
from typing import Sequence
from typing import Union

import numpy as np


Dtype = Union[Callable, type, str]


# valid_features_str_names = []


class ColumnRole:
    """Abstract class for column role.

    Role type defines column dtype,
    place of column in dataset and transformers
    and set additional attributes which impacts
    on the way how it's handled.

    """

    dtype = object
    force_input = False
    _name = "Abstract"

    @property
    def name(self) -> str:
        """Get str role name.

        Returns:
            str role name.

        """
        return self._name

    def __repr__(self) -> str:
        """String view of role.

        Returns:
            Representation string.

        """
        params = [(x, self.__dict__[x]) for x in self.__dict__ if x not in ["dtype", "name"]]

        return "{0} role, dtype {1}. Additional params: {2}".format(self.name, self.dtype, params)

    def __hash__(self) -> int:
        """Define how to hash - hash from str view.

        Returns:
            Hashed name of column.

        """
        return hash(self.__repr__())

    def __eq__(self, other: Any) -> bool:
        """Define how to compare - if reprs are equal (hashed).

        Args:
            other: Another :class:`~lightautoml.dataset.roles.ColumnRole`.

        Returns:
            ``True`` if equal.

        """
        return self.__repr__() == other.__repr__()

    @staticmethod
    def from_string(name: str, **kwargs: Any) -> "ColumnRole":
        """Create default params role from string.

        Args:
            name: Role name.
            kwargs: Other parameters.

        Returns:
            Corresponding role object.

        """
        name = name.lower()

        if name in ["target"]:
            return TargetRole(**kwargs)

        if name in ["numeric"]:
            return NumericRole(**kwargs)

        if name in ["category"]:
            return CategoryRole(**kwargs)

        if name in ["text"]:
            return TextRole(**kwargs)

        if name in ["datetime"]:
            return DatetimeRole(**kwargs)

        if name in ["base_date"]:
            kwargs = {**{"seasonality": (), "base_date": True}, **kwargs}
            return DatetimeRole(**kwargs)

        if name in ["date"]:
            kwargs = {**{"seasonality": (), "base_date": True}, **kwargs}
            return DateRole(**kwargs)

        if name in ["group"]:
            return GroupRole()

        if name in ["drop"]:
            return DropRole()

        if name in ["id"]:
            # kwargs = {**{"encoding_type": "oof", "unknown": 1}, **kwargs}
            return IdRole()

        if name in ["folds"]:
            return FoldsRole()

        if name in ["weights"]:
            return WeightsRole()

        if name in ["path"]:
            return PathRole()

        if name in ["treatment"]:
            return TreatmentRole()

        raise ValueError("Unknown string role: {}".format(name))


class NumericRole(ColumnRole):
    """Numeric role.

    Args:
        dtype: Variable type.
        force_input: Select a feature for training,
            regardless of the selector results.
        prob: If input number is probability.
        discretization: Flag of discretization.

    """

    _name = "Numeric"

    def __init__(
        self,
        dtype: Dtype = np.float32,
        force_input: bool = False,
        prob: bool = False,
        discretization: bool = False,
    ):
        self.dtype = dtype
        self.force_input = force_input
        self.prob = prob
        self.discretization = discretization


class CategoryRole(ColumnRole):
    """Category role.

    Args:
        dtype: Variable type.
        encoding_type: Encoding type.
        unknown: Cut-off freq to process rare categories as unseen.
        force_input: Select a feature for training,
            regardless of the selector results.
        ordinal: Ordinal category.

    Note:
        Valid encoding_type:

            - `'auto'` - default processing
            - `'int'` - encode with int
            - `'oof'` - out-of-fold target encoding
            - `'freq'` - frequency encoding
            - `'ohe'` - one hot encoding

    """

    _name = "Category"

    def __init__(
        self,
        dtype: Dtype = object,
        encoding_type: str = "auto",
        unknown: int = 5,
        force_input: bool = False,
        label_encoded: bool = False,
        ordinal: bool = False,
    ):
        # TODO: assert dtype is object, 'Dtype for category should be defined' ?
        # assert encoding_type == 'auto', 'For the moment only auto is supported'
        # TODO: support all encodings
        self.dtype = dtype
        self.encoding_type = encoding_type
        self.unknown = unknown
        self.force_input = force_input
        self.label_encoded = label_encoded
        self.ordinal = ordinal


class TextRole(ColumnRole):
    """Text role.

    Args:
        dtype: Variable type.
        force_input: Select a feature for training,
            regardless of the selector results.

    """

    _name = "Text"

    def __init__(self, dtype: Dtype = str, force_input: bool = True):
        self.dtype = dtype
        self.force_input = force_input


class DatetimeRole(ColumnRole):
    """Datetime role.

    Args:
        dtype: Variable type.
        seasonality: Seasons to extract from date.
            Valid are: 'y', 'm', 'd', 'wd', 'hour', 'min', 'sec', 'ms', 'ns'.
        base_date: Base date is used to calculate difference
            with other dates, like `age = report_dt - birth_dt`.
        date_format: Format to parse date.
        unit: The unit of the arg denote the unit, pandas like, see more:
            https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.to_datetime.html.
        origin: Define the reference date, pandas like, see more:
            https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.to_datetime.html.
        force_input: Select a feature for training,
            regardless of the selector results.
        base_feats: To calculate feats on base date.
        country: Datetime metadata to extract holidays.
        prov: Datetime metadata to extract holidays.
        state: Datetime metadata to extract holidays.

    """

    _name = "Datetime"

    def __init__(
        self,
        dtype: Dtype = np.datetime64,
        seasonality: Optional[Sequence[str]] = ("y", "m", "wd"),
        base_date: bool = False,
        date_format: Optional[str] = None,
        unit: Optional[str] = None,
        origin: Union[str, datetime] = "unix",
        force_input: bool = False,
        base_feats: bool = True,
        country: Optional[str] = None,
        prov: Optional[str] = None,
        state: Optional[str] = None,
    ):
        self.dtype = dtype
        self.seasonality = []
        if seasonality is not None:
            self.seasonality = seasonality
        self.base_date = base_date
        self.format = date_format
        self.unit = unit
        self.origin = origin

        self.force_input = force_input
        if self.base_date:
            self.force_input = True
        self.base_feats = base_feats

        self.country = country
        self.prov = prov
        self.state = state


# class MixedRole(ColumnRole):
#     """
#     Mixed role. If exact role extraction is difficult, it goes into both pipelines
#     """


class TargetRole(ColumnRole):
    """Target role.

    Args:
        dtype: Dtype of target.

    """

    _name = "Target"

    def __init__(self, dtype: Dtype = np.float32):
        self.dtype = dtype


class GroupRole(ColumnRole):
    """Group role."""

    _name = "Group"


class DropRole(ColumnRole):
    """Drop role."""

    _name = "Drop"


class WeightsRole(ColumnRole):
    """Weights role."""

    _name = "Weights"


class FoldsRole(ColumnRole):
    """Folds role."""

    _name = "Folds"


class PathRole(ColumnRole):
    """Path role."""

    _name = "Path"


class DateRole(DatetimeRole):
    """Date role."""

    _name = "Date"


class IdRole(CategoryRole):
    """Id role."""

    _name = "Id"


class TreatmentRole(ColumnRole):
    """Uplift Treatment Role."""

    _name = "Treatment"



import logging
import os
import sys


_root_logger = logging.getLogger()
_logger = logging.getLogger(__name__)
_logger.setLevel(logging.WARNING)

# if root logger has handlers, propagate messages up and let root logger process them
if not _root_logger.hasHandlers():
    _logger.addHandler(logging.StreamHandler(sys.stdout))
    _logger.propagate = False

__all__ = [
    "automl",
    "dataset",
    "ml_algo",
    "pipelines",
    "image",
    "reader",
    "transformers",
    "validation",
    "text",
    "tasks",
    "utils",
    "addons",
    "report",
]

if os.getenv("DOCUMENTATION_ENV") is None:
    try:
        import importlib.metadata as importlib_metadata
    except ModuleNotFoundError:
        import importlib_metadata

    __version__ = importlib_metadata.version(__name__)



"""Internal representation of dataset in numpy, pandas and csr formats."""

from copy import copy  # , deepcopy
from typing import Any
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import TypeVar
from typing import Union

import numpy as np
import pandas as pd

from pandas import DataFrame
from pandas import Series
from scipy import sparse

from ..tasks.base import Task
from .base import IntIdx
from .base import LAMLDataset
from .base import RolesDict
from .base import array_attr_roles
from .base import valid_array_attributes
from .roles import ColumnRole
from .roles import DropRole
from .roles import NumericRole


# disable warnings later
# pd.set_option('mode.chained_assignment', None)

NpFeatures = Union[Sequence[str], str, None]
NpRoles = Union[Sequence[ColumnRole], ColumnRole, RolesDict, None]
DenseSparseArray = Union[np.ndarray, sparse.csr_matrix]
FrameOrSeries = Union[DataFrame, Series]
Dataset = TypeVar("Dataset", bound=LAMLDataset)


# possible checks list
# valid shapes
# target var is ok for task
# pandas - roles for all columns are defined
# numpy - roles and features are ok each other
# numpy - roles and features are ok for data
# features names does not contain __ - it's used to split processing names

# sparse - do not replace init and set data, but move type assert in checks?


class NumpyDataset(LAMLDataset):
    """Dataset that contains info in np.ndarray format.

    Create dataset from numpy arrays.

    Args:
        data: 2d array of features.
        features: Features names.
        roles: Roles specifier.
        task: Task specifier.
        **kwargs: Named attributes like target, group etc ..

    Note:
        For different type of parameter feature there is different behavior:

            - list, should be same len as data.shape[1]
            - None - automatic set names like feat_0, feat_1 ...
            - Prefix - automatic set names like Prefix_0, Prefix_1 ...

        For different type of parameter feature there is different behavior:

            - list, should be same len as data.shape[1].
            - None - automatic set NumericRole(np.float32).
            - ColumnRole - single role.
            - dict.

    """

    # TODO: Checks here
    _init_checks = ()
    _data_checks = ()
    _concat_checks = ()
    _dataset_type = "NumpyDataset"

    @property
    def features(self) -> List[str]:
        """Features list."""
        return list(self._features)

    @features.setter
    def features(self, val: Union[Sequence[str], str, None]):
        """Define how to set features.

        Args:
            val: Values of features.

        Note:
            There is different behavior for different type of val parameter:

                - list - should be same len as ``data.shape[1]``
                - None - automatic set names like `feat_0`, `feat_1` ...
                - `'Prefix'` - automatic set names
                  to `Prefix_0`, `Prefix_1` ...

        """
        if type(val) is list:
            self._features = copy(val)
        else:
            prefix = val if val is not None else "feat"
            self._features = ["{0}_{1}".format(prefix, x) for x in range(self.data.shape[1])]

    @property
    def roles(self) -> RolesDict:
        """Roles dict."""
        return copy(self._roles)

    @roles.setter
    def roles(self, val: NpRoles):
        """Define how to set roles.

        Args:
            val: Roles.

        Note:
            There is different behavior for different type of val parameter:

                - `List` - should be same len as ``data.shape[1]``.
                - `None` - automatic set ``NumericRole(np.float32)``.
                - ``ColumnRole`` - single role for all.
                - ``dict``.

        """
        if type(val) is dict:
            self._roles = dict(((x, val[x]) for x in self.features))
        elif type(val) is list:
            self._roles = dict(zip(self.features, val))
        else:
            role = NumericRole(np.float32) if val is None else val
            self._roles = dict(((x, role) for x in self.features))

    def _check_dtype(self):
        """Check if dtype in ``.set_data`` is ok and cast if not.

        Raises:
            AttributeError: If there is non-numeric type in dataset.

        """
        # dtypes = list(set(map(lambda x: x.dtype, self.roles.values())))
        dtypes = list(set([i.dtype for i in self.roles.values()]))
        self.dtype = np.find_common_type(dtypes, [])

        for f in self.roles:
            self._roles[f].dtype = self.dtype

        assert np.issubdtype(self.dtype, np.number), "Support only numeric types in numpy dataset."

        if self.data.dtype != self.dtype:
            try:
                self.data = self.data.astype(self.dtype)
            except:
                pass

    def __init__(
        self,
        data: Optional[DenseSparseArray],
        features: NpFeatures = (),
        roles: NpRoles = None,
        task: Optional[Task] = None,
        **kwargs: np.ndarray
    ):

        self._initialize(task, **kwargs)
        if data is not None:
            self.set_data(data, features, roles)

    def set_data(self, data: DenseSparseArray, features: NpFeatures = (), roles: NpRoles = None):
        """Inplace set data, features, roles for empty dataset.

        Args:
            data: 2d np.ndarray of features.
            features: features names.
            roles: Roles specifier.

        Note:
            For different type of parameter feature there is different behavior:

                - List, should be same len as data.shape[1]
                - None - automatic set names like feat_0, feat_1 ...
                - Prefix - automatic set names like Prefix_0, Prefix_1 ...

            For different type of parameter feature there is different behavior:

                - List, should be same len as data.shape[1].
                - None - automatic set NumericRole(np.float32).
                - ColumnRole - single role.
                - dict.

        """
        assert data is None or type(data) is np.ndarray, "Numpy dataset support only np.ndarray features"
        super().set_data(data, features, roles)
        self._check_dtype()

    @staticmethod
    def _hstack(datasets: Sequence[np.ndarray]) -> np.ndarray:
        """Concatenate function for numpy arrays.

        Args:
            datasets: Sequence of np.ndarray.

        Returns:
            Stacked features array.

        """
        return np.hstack(datasets)

    @staticmethod
    def _get_rows(data: np.ndarray, k: IntIdx) -> np.ndarray:
        """Get rows slice for numpy ndarray.

        Args:
            data: Data.
            k: Sequence of integer indexes.

        Returns:
            Rows slice.

        """
        return data[k]

    @staticmethod
    def _get_cols(data: np.ndarray, k: IntIdx) -> np.ndarray:
        """Get cols slice.

        Args:
            data: Data.
            k: Sequence of integer indexes.

        Returns:
            Cols slice.

        """
        return data[:, k]

    @classmethod
    def _get_2d(cls, data: np.ndarray, k: Tuple[IntIdx, IntIdx]) -> np.ndarray:
        """Get 2d slice.

        Args:
            data: Data.
            k: Tuple of integer sequences.

        Returns:
            2d slice.

        """
        rows, cols = k

        return data[rows, cols]

    @staticmethod
    def _set_col(data: np.ndarray, k: int, val: np.ndarray):
        """Inplace set columns.

        Args:
            data: Data.
            k: Index of column.
            val: Values to set.

        """
        data[:, k] = val

    def to_numpy(self) -> "NumpyDataset":
        """Empty method to convert to numpy.

        Returns:
            Same NumpyDataset.

        """
        return self

    def to_csr(self) -> "CSRSparseDataset":
        """Convert to csr.

        Returns:
            Same dataset in CSRSparseDatatset format.

        """
        assert all(
            [self.roles[x].name == "Numeric" for x in self.features]
        ), "Only numeric data accepted in sparse dataset"
        data = None if self.data is None else sparse.csr_matrix(self.data)

        roles = self.roles
        features = self.features
        # target and etc ..
        params = dict(((x, self.__dict__[x]) for x in self._array_like_attrs))
        task = self.task

        return CSRSparseDataset(data, features, roles, task, **params)

    def to_pandas(self) -> "PandasDataset":
        """Convert to PandasDataset.

        Returns:
            Same dataset in PandasDataset format.

        """
        # check for empty case
        data = None if self.data is None else DataFrame(self.data, columns=self.features)
        roles = self.roles
        # target and etc ..
        params = dict(
            (
                (x, Series(self.__dict__[x]) if len(self.__dict__[x].shape) == 1 else DataFrame(self.__dict__[x]))
                for x in self._array_like_attrs
            )
        )
        task = self.task

        return PandasDataset(data, roles, task, **params)

    @staticmethod
    def from_dataset(dataset: Dataset) -> "NumpyDataset":
        """Convert random dataset to numpy.

        Args:
            dataset: Dataset.

        Returns:
            Numpy dataset.

        """
        return dataset.to_numpy()


class CSRSparseDataset(NumpyDataset):
    """Dataset that contains sparse features and np.ndarray targets."""

    _init_checks = ()
    _data_checks = ()
    _concat_checks = ()
    _dataset_type = "CSRSparseDataset"

    @staticmethod
    def _get_cols(data: Any, k: Any):
        """Not implemented."""
        raise NotImplementedError

    @staticmethod
    def _set_col(data: Any, k: Any, val: Any):
        """Not implemented."""
        raise NotImplementedError

    def to_pandas(self) -> Any:
        """Not implemented."""
        raise NotImplementedError

    def to_numpy(self) -> "NumpyDataset":
        """Convert to NumpyDataset.

        Returns:
            NumpyDataset.

        """
        # check for empty
        data = None if self.data is None else self.data.toarray()
        roles = self.roles
        features = self.features
        # target and etc ..
        params = dict(((x, self.__dict__[x]) for x in self._array_like_attrs))
        task = self.task

        return NumpyDataset(data, features, roles, task, **params)

    @property
    def shape(self) -> Tuple[Optional[int], Optional[int]]:
        """Get size of 2d feature matrix.

        Returns:
            tuple of 2 elements.

        """
        rows, cols = None, None
        try:
            rows, cols = self.data.shape
        except TypeError:
            if len(self._array_like_attrs) > 0:
                rows = len(self.__dict__[self._array_like_attrs[0]])
        return rows, cols

    @staticmethod
    def _hstack(datasets: Sequence[Union[sparse.csr_matrix, np.ndarray]]) -> sparse.csr_matrix:
        """Concatenate function for sparse and numpy arrays.

        Args:
            datasets: Sequence of csr_matrix or np.ndarray.

        Returns:
            Sparse matrix.

        """
        return sparse.hstack(datasets, format="csr")

    def __init__(
        self,
        data: Optional[DenseSparseArray],
        features: NpFeatures = (),
        roles: NpRoles = None,
        task: Optional[Task] = None,
        **kwargs: np.ndarray
    ):
        """Create dataset from csr_matrix.

        Args:
            data: csr_matrix of features.
            features: Features names.
            roles: Roles specifier.
            task: Task specifier.
            **kwargs: Named attributes like target, group etc ..

        Note:
            For different type of parameter feature there is different behavior:

                - list, should be same len as data.shape[1]
                - None - automatic set names like feat_0, feat_1 ...
                - Prefix - automatic set names like Prefix_0, Prefix_1 ...

            For different type of parameter feature there is different behavior:

                - list, should be same len as data.shape[1].
                - None - automatic set NumericRole(np.float32).
                - ColumnRole - single role.
                - dict.

        """
        self._initialize(task, **kwargs)
        if data is not None:
            self.set_data(data, features, roles)

    def set_data(self, data: DenseSparseArray, features: NpFeatures = (), roles: NpRoles = None):
        """Inplace set data, features, roles for empty dataset.

        Args:
            data: csr_matrix of features.
            features: features names.
            roles: Roles specifier.

        Note:
            For different type of parameter feature there is different behavior:

                - list, should be same len as data.shape[1]
                - None - automatic set names like feat_0, feat_1 ...
                - Prefix - automatic set names like Prefix_0, Prefix_1 ...

            For different type of parameter feature there is different behavior:

                - list, should be same len as data.shape[1].
                - None - automatic set NumericRole(np.float32).
                - ColumnRole - single role.
                - dict.

        """
        assert data is None or type(data) is sparse.csr_matrix, "CSRSparseDataset support only csr_matrix features"
        LAMLDataset.set_data(self, data, features, roles)
        self._check_dtype()

    @staticmethod
    def from_dataset(dataset: Dataset) -> "CSRSparseDataset":
        """Convert dataset to sparse dataset.

        Args:
            dataset: Dataset.

        Returns:
            Dataset in sparse form.

        """
        return dataset.to_csr()


class PandasDataset(LAMLDataset):
    """Dataset that contains `pd.DataFrame` features and `pd.Series` targets.

    Args:
        data: Table with features.
        features: features names.
        roles: Roles specifier.
        task: Task specifier.
        **kwargs: Series, array like attrs target, group etc...

    """

    _init_checks = ()
    _data_checks = ()
    _concat_checks = ()
    _dataset_type = "PandasDataset"

    @property
    def features(self) -> List[str]:
        """Get list of features.

        Returns:
            list of features.

        """
        return [] if self.data is None else list(self.data.columns)

    @features.setter
    def features(self, val: None):
        """Ignore setting features.

        Args:
            val: ignored.

        """
        pass

    def __init__(
        self,
        data: Optional[DataFrame] = None,
        roles: Optional[RolesDict] = None,
        task: Optional[Task] = None,
        **kwargs: Series
    ):
        if roles is None:
            roles = {}
        # parse parameters
        # check if target, group etc .. defined in roles
        for f in roles:
            for k, r in zip(valid_array_attributes, array_attr_roles):
                if roles[f].name == r:
                    kwargs[k] = data[f].reset_index(drop=True)
                    roles[f] = DropRole()
        self._initialize(task, **kwargs)
        if data is not None:
            self.set_data(data, None, roles)

    def _get_cols_idx(self, columns: Union[Sequence[str], str]) -> Union[Sequence[int], int]:
        """Get numeric index of columns by column names.

        Args:
            columns: sequence of columns of single column.

        Returns:
            sequence of int indexes or single int.

        """
        if isinstance(columns, str):
            idx = self.data.columns.get_loc(columns)

        else:
            idx = self.data.columns.get_indexer(columns)

        return idx

    def set_data(self, data: DataFrame, features: None, roles: RolesDict):
        """Inplace set data, features, roles for empty dataset.

        Args:
            data: Table with features.
            features: `None`, just for same interface.
            roles: Dict with roles.

        """
        super().set_data(data, features, roles)
        self._check_dtype()

    def _check_dtype(self):
        """Check if dtype in .set_data is ok and cast if not."""
        date_columns = []

        self.dtypes = {}
        for f in self.roles:
            if self.roles[f].name == "Datetime":
                date_columns.append(f)
            else:
                self.dtypes[f] = self.roles[f].dtype

        self.data = self.data.astype(self.dtypes)
        self.data.reset_index(drop=True, inplace=True)
        # do we need to reset_index ?? If yes - drop for Series attrs too
        # case to check - concat pandas dataset and from numpy to pandas dataset
        # TODO: Think about reset_index here
        # self.data.reset_index(inplace=True, drop=True)

        # handle dates types
        for i in date_columns:
            dt_role = self.roles[i]
            if not (self.data.dtypes[i] is np.datetime64):
                self.data[i] = pd.to_datetime(
                    self.data[i],
                    format=dt_role.format,
                    unit=dt_role.unit,
                    origin=dt_role.origin,
                    cache=True,
                )

            self.dtypes[i] = np.datetime64

    @staticmethod
    def _hstack(datasets: Sequence[DataFrame]) -> DataFrame:
        """Define how to concat features arrays.

        Args:
            datasets: Sequence of tables.

        Returns:
            concatenated table.

        """
        return pd.concat(datasets, axis=1)

    @staticmethod
    def _get_rows(data: DataFrame, k: IntIdx) -> FrameOrSeries:
        """Define how to get rows slice.

        Args:
            data: Table with data.
            k: Sequence of `int` indexes or `int`.

        Returns:
            Sliced rows.

        """
        return data.iloc[k]

    @staticmethod
    def _get_cols(data: DataFrame, k: IntIdx) -> FrameOrSeries:
        """Define how to get cols slice.

        Args:
            data: Table with data.
            k: Sequence of `int` indexes or `int`.

        Returns:
            Sliced cols.

        """
        return data.iloc[:, k]

    @classmethod
    def _get_2d(cls, data: DataFrame, k: Tuple[IntIdx, IntIdx]) -> FrameOrSeries:
        """Define 2d slice of table.

        Args:
            data: Table with data.
            k: Sequence of `int` indexes or `int`.

        Returns:
            2d sliced table.

        """
        rows, cols = k

        return data.iloc[rows, cols]

    @staticmethod
    def _set_col(data: DataFrame, k: int, val: Union[Series, np.ndarray]):
        """Inplace set column value to `pd.DataFrame`.

        Args:
            data: Table with data.
            k: Column index.
            val: Values to set.

        """
        data.iloc[:, k] = val

    def to_numpy(self) -> "NumpyDataset":
        """Convert to class:`NumpyDataset`.

        Returns:
            Same dataset in class:`NumpyDataset` format.

        """
        # check for empty
        data = None if self.data is None else self.data.values
        roles = self.roles
        features = self.features
        # target and etc ..
        params = dict(((x, self.__dict__[x].values) for x in self._array_like_attrs))
        task = self.task

        return NumpyDataset(data, features, roles, task, **params)

    def to_pandas(self) -> "PandasDataset":
        """Empty method, return the same object.

        Returns:
            Self.

        """
        return self

    @staticmethod
    def from_dataset(dataset: Dataset) -> "PandasDataset":
        """Convert random dataset to pandas dataset.

        Args:
            dataset: Dataset.

        Returns:
            Converted to pandas dataset.

        """
        return dataset.to_pandas()

    def nan_rate(self):
        """Counts overall number of nans in dataset.

        Returns:
            Number of nans.

        """
        return (len(self.data) - self.data.count()).sum()



"""Tokenizer classes for text preprocesessing and tokenization."""

import re

from functools import partial
from multiprocessing import Pool
from typing import Any
from typing import List
from typing import Optional
from typing import Sequence
from typing import Union


try:
    import nltk

    from nltk.stem import SnowballStemmer
except:
    import warnings

    warnings.warn("'nltk' - package isn't installed")


from ..dataset.base import RolesDict
from ..dataset.roles import ColumnRole


Roles = Union[Sequence[ColumnRole], ColumnRole, RolesDict, None]


def tokenizer_func(arr, tokenizer):
    """Additional tokenizer function."""
    return [tokenizer._tokenize(x) for x in arr]


class BaseTokenizer:
    """Base class for tokenizer method.

    Tokenization with simple text cleaning and preprocessing.

    Args:
        n_jobs: Number of threads for multiprocessing.
        to_string: Return string or list of tokens.
        **kwargs: Ignore.

    """

    _fname_prefix = None
    _fit_checks = ()
    _transform_checks = ()

    def __init__(self, n_jobs: int = 4, to_string: bool = True, **kwargs: Any):
        self.n_jobs = n_jobs
        self.to_string = to_string

    def preprocess_sentence(self, snt: str) -> str:
        """Preprocess sentence string (lowercase, etc.).

        Args:
            snt: Sentence string.

        Returns:
            Resulting string.

        """
        return snt

    def tokenize_sentence(self, snt: str) -> List[str]:
        """Convert sentence string to a list of tokens.

        Args:
            snt: Sentence string.

        Returns:
            Resulting list of tokens.

        """
        return snt.split(" ")

    def filter_tokens(self, snt: List[str]) -> List[str]:
        """Clean list of sentence tokens.

        Args:
            snt: List of tokens.

        Returns:
            Resulting list of filtered tokens

        """
        return snt

    def postprocess_tokens(self, snt: List[str]) -> List[str]:
        """Additional processing steps: lemmatization, pos tagging, etc.

        Args:
            snt: List of tokens.

        Returns:
            Resulting list of processed tokens.

        """
        return snt

    def postprocess_sentence(self, snt: str) -> str:
        """Postprocess sentence string (merge words).

        Args:
            snt: Sentence string.

        Returns:
            Resulting string.

        """
        return snt

    def _tokenize(self, snt: str) -> Union[List[str], str]:
        """Tokenize text string.

        Args:
            snt: String.

        Returns:
            Resulting tokenized list.

        """
        res = self.preprocess_sentence(snt)
        res = self.tokenize_sentence(res)
        res = self.filter_tokens(res)
        res = self.postprocess_tokens(res)

        if self.to_string:
            res = " ".join(res)
            res = self.postprocess_sentence(res)
        return res

    def tokenize(self, text: List[str]) -> Union[List[List[str]], List[str]]:
        """Tokenize list of texts.

        Args:
            text: List of texts.

        Returns:
            Resulting tokenized list.

        """
        if self.n_jobs == 1:
            res = self._tokenize_singleproc(text)
        else:
            res = self._tokenize_multiproc(text)
        return res

    def _tokenize_singleproc(self, snt: List[str]) -> List[str]:
        """Singleproc version of tokenization.

        Args:
            snt: List of texts.

        Returns:
            List of tokenized texts.

        """
        return [self._tokenize(x) for x in snt]

    def _tokenize_multiproc(self, snt: List[str]) -> List[str]:
        """Multiproc version of tokenization.

        Args:
            snt: List of texts.

        Returns:
            List of tokenized texts.

        """
        idx = list(range(0, len(snt), len(snt) // self.n_jobs + 1)) + [len(snt)]

        parts = [snt[i:j] for (i, j) in zip(idx[:-1], idx[1:])]

        f = partial(tokenizer_func, tokenizer=self)
        with Pool(self.n_jobs) as p:
            res = p.map(f, parts)
        del f

        tokens = res[0]
        for r in res[1:]:
            tokens.extend(r)
        return tokens


class SimpleRuTokenizer(BaseTokenizer):
    """Russian tokenizer.

    Include numeric, punctuation and short word filtering.
    Use stemmer by default and do lowercase.

    Args:
        n_jobs: Number of threads for multiprocessing.
        to_string: Return string or list of tokens.
        stopwords: Use stopwords or not.
        is_stemmer: Use stemmer.
        kwargs: Ignore.

    """

    def __init__(
        self,
        n_jobs: int = 4,
        to_string: bool = True,
        stopwords: Optional[Union[bool, Sequence[str]]] = False,
        is_stemmer: bool = True,
        **kwargs: Any
    ):
        super().__init__(n_jobs, **kwargs)
        self.n_jobs = n_jobs
        self.to_string = to_string
        if isinstance(stopwords, (tuple, list, set)):
            self.stopwords = set(stopwords)
        elif stopwords:
            self.stopwords = set(nltk.corpus.stopwords.words("russian"))
        else:
            self.stopwords = {}

        self.stemmer = SnowballStemmer("russian", ignore_stopwords=len(self.stopwords) > 0) if is_stemmer else None

    @staticmethod
    def _is_abbr(word: str) -> bool:
        """Check if the word is an abbreviation."""
        return sum([x.isupper() and x.isalpha() for x in word]) > 1 and len(word) <= 5

    def preprocess_sentence(self, snt: str) -> str:
        """Preprocess sentence string (lowercase, etc.).

        Args:
            snt: Sentence string.

        Returns:
            Resulting string.

        """
        snt = snt.strip()
        snt = snt.replace("Ё", "Е").replace("ё", "е")
        s = re.sub("[^A-Za-zА-Яа-я0-9]+", " ", snt)
        s = re.sub(r"^\d+\s|\s\d+\s|\s\d+$", " ", s)
        return s

    def tokenize_sentence(self, snt: str) -> List[str]:
        """Convert sentence string to a list of tokens.

        Args:
            snt: Sentence string.

        Returns:
            Resulting list of tokens.

        """
        return snt.split(" ")

    def filter_tokens(self, snt: List[str]) -> List[str]:
        """Clean list of sentence tokens.

        Args:
            snt: List of tokens.

        Returns:
            Resulting list of filtered tokens.

        """
        filtered_s = []
        for w in snt:

            # ignore numbers
            if w.isdigit():
                pass
            elif w.lower() in self.stopwords:
                pass
            elif self._is_abbr(w):
                filtered_s.append(w)
            # ignore short words
            elif len(w) < 2:
                pass
            elif w.isalpha():
                filtered_s.append(w.lower())
        return filtered_s

    def postprocess_tokens(self, snt: List[str]) -> List[str]:
        """Additional processing steps: lemmatization, pos tagging, etc.

        Args:
            snt: List of tokens.

        Returns:
            Resulting list of processed tokens.

        """
        if self.stemmer is not None:
            return [self.stemmer.stem(w.lower()) for w in snt]
        else:
            return snt

    def postprocess_sentence(self, snt: str) -> str:
        """Postprocess sentence string (merge words).

        Args:
            snt: Sentence string.

        Returns:
            Resulting string.

        """
        snt = (" " + snt).replace(" не ", " не")
        snt = snt.replace(" ни ", " ни")
        return snt[1:]


class SimpleEnTokenizer(BaseTokenizer):
    """English tokenizer.

    Args:
        n_jobs: Number of threads for multiprocessing.
        to_string: Return string or list of tokens.
        stopwords: Use stopwords or not.
        is_stemmer: Use stemmer.
        kwargs: Ignore.

    """

    def __init__(
        self,
        n_jobs: int = 4,
        to_string: bool = True,
        stopwords: Optional[Union[bool, Sequence[str]]] = False,
        is_stemmer: bool = True,
        **kwargs: Any
    ):
        super().__init__(n_jobs, **kwargs)
        self.n_jobs = n_jobs
        self.to_string = to_string
        if isinstance(stopwords, (tuple, list, set)):
            self.stopwords = set(stopwords)
        elif stopwords:
            self.stopwords = set(nltk.corpus.stopwords.words("english"))
        else:
            self.stopwords = {}

        self.stemmer = SnowballStemmer("english", ignore_stopwords=len(self.stopwords) > 0) if is_stemmer else None

    def preprocess_sentence(self, snt: str) -> str:
        """Preprocess sentence string (lowercase, etc.).

        Args:
            snt: Sentence string.

        Returns:
            Resulting string.

        """
        snt = snt.strip()
        s = re.sub("[^A-Za-zА-Яа-я0-9]+", " ", snt)
        s = re.sub(r"^\d+\s|\s\d+\s|\s\d+$", " ", s)
        return s

    def tokenize_sentence(self, snt: str) -> List[str]:
        """Convert sentence string to a list of tokens.

        Args:
            snt: Sentence string.

        Returns:
            Resulting list of tokens.

        """
        return snt.split(" ")

    def filter_tokens(self, snt: List[str]) -> List[str]:
        """Clean list of sentence tokens.

        Args:
            snt: List of tokens.

        Returns:
            Resulting list of filtered tokens.

        """
        if len(self.stopwords) > 0:
            filtered_s = []
            for w in snt:
                if w.lower() not in self.stopwords:
                    filtered_s.append(w)
            return filtered_s
        else:
            return snt

    def postprocess_tokens(self, snt: List[str]) -> List[str]:
        """Additional processing steps: lemmatization, pos tagging, etc.

        Args:
            snt: List of tokens.

        Returns:
            Resulting list of processed tokens.

        """
        if self.stemmer is not None:
            return [self.stemmer.stem(w.lower()) for w in snt]
        else:
            return snt

    def postprocess_sentence(self, snt: str) -> str:
        """Postprocess sentence string (merge words).

        Args:
            snt: Sentence string.

        Returns:
            Resulting string.

        """
        return snt



"""Text utility script."""

import os
import random

from typing import Dict
from typing import List
from typing import Sequence

import numpy as np
import torch

from sklearn.utils.murmurhash import murmurhash3_32


_dtypes_mapping = {
    "label": "float",
    "cat": "long",
    "cont": "float",
    "weight": "float",
    "input_ids": "long",
    "attention_mask": "long",
    "token_type_ids": "long",
    "text": "float",  # embeddings
    "length": "long",
}


def inv_sigmoid(x: np.ndarray) -> np.ndarray:
    """Inverse sigmoid transformation.

    Args:
        x: Input array.

    Returns:
        Transformed array.

    """
    return np.log(x / (1 - x))


def inv_softmax(x: np.ndarray) -> np.ndarray:
    """Variant of inverse softmax transformation with zero constant term.

    Args:
        x: Input array.

    Returns:
        Transformed array.

    """
    eps = 1e-7
    x = np.abs(x)
    arr = (x + eps) / (np.sum(x) + eps)
    arr = np.log(arr)
    return arr


def is_shuffle(stage: str) -> bool:
    """Whether shuffle input.

    Args:
        stage: Train, val, test.

    Returns:
        Bool value.

    """
    is_sh = {"train": True, "val": False, "test": False}
    return is_sh[stage]


def seed_everything(seed: int = 42, deterministic: bool = True):
    """Set random seed and cudnn params.

    Args:
        seed: Random state.
        deterministic: cudnn backend.

    """
    random.seed(seed)
    os.environ["PYTHONHASHSEED"] = str(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    if deterministic:
        torch.backends.cudnn.deterministic = True


def parse_devices(dvs, is_dp: bool = False) -> tuple:
    """Parse devices and convert first to the torch device.

    Args:
        dvs: List, string with device ids or torch.device.
        is_dp: Use data parallel - additionally returns device ids.

    Returns:
        First torch device and list of gpu ids.

    """
    device = []
    ids = []
    if (not torch.cuda.is_available()) or (dvs is None):
        return torch.device("cpu"), None

    if not isinstance(dvs, (list, tuple)):
        dvs = [dvs]

    for _device in dvs:
        if isinstance(_device, str):
            if _device.startswith("cuda:"):
                ids.append(int(_device.split("cuda:")[-1]))
            elif _device == "cuda":
                ids.append(0)
            elif _device == "cpu":
                return torch.device("cpu"), None
            else:
                ids.append(int(_device))
                _device = torch.device(int(_device))

        elif isinstance(_device, int):
            ids.append(_device)
            _device = torch.device("cuda:{}".format(_device))
        elif isinstance(_device, torch.device):
            if _device.type == "cpu":
                return _device, None
            else:
                if _device.index is None:
                    ids.append(0)
                else:
                    ids.append(_device.index)
        else:
            raise ValueError("Unknown device type: {}".format(_device))

        device.append(_device)

    return device[0], ids if (len(device) > 1) and is_dp else None


def custom_collate(batch: List[np.ndarray]) -> torch.Tensor:
    """Puts each data field into a tensor with outer dimension batch size."""
    elem = batch[0]
    if isinstance(elem, torch.Tensor):
        out = None
        numel = sum([x.numel() for x in batch])
        storage = elem.storage()._new_shared(numel)
        out = elem.new(storage)
        return torch.stack(batch, 0, out=out)
    else:
        return torch.from_numpy(np.array(batch)).float()


def collate_dict(batch: List[Dict[str, np.ndarray]]) -> Dict[str, torch.Tensor]:
    """custom_collate for dicts."""
    keys = list(batch[0].keys())
    transposed_data = list(map(list, zip(*[tuple([i[name] for name in i.keys()]) for i in batch])))
    return {key: custom_collate(transposed_data[n]) for n, key in enumerate(keys)}


def single_text_hash(x: str) -> str:
    """Get text hash.

    Args:
        x: Text.

    Returns:
        String text hash.

    """
    numhash = murmurhash3_32(x, seed=13)
    texthash = str(numhash) if numhash > 0 else "m" + str(abs(numhash))
    return texthash


def get_textarr_hash(x: Sequence[str]) -> str:
    """Get hash of array with texts.

    Args:
        x: Text array.

    Returns:
        Hash of array.

    """
    full_hash = single_text_hash(str(x))
    n = 0
    for text in x:
        if text != "":
            full_hash += "_" + single_text_hash(text)
            n += 1
            if n >= 3:
                break

    return full_hash



"""Contains base classes for internal dataset interface."""

from copy import copy  # , deepcopy
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

from ..tasks.base import Task
from .roles import ColumnRole


valid_array_attributes = ("target", "group", "folds", "weights", "date", "id", "treatment")
array_attr_roles = ("Target", "Group", "Folds", "Weights", "Date", "Id", "Treatment")
# valid_tasks = ('reg', 'binary', 'multiclass') # TODO: Add multiclass and multilabel. Refactor for some dataset and pipes needed
# valid_tasks = ('reg', 'binary')


RolesDict = Dict[str, ColumnRole]
IntIdx = Union[Sequence[int], int]
RowSlice = Optional[Union[Sequence[int], Sequence[bool]]]
ColSlice = Optional[Union[Sequence[str], str]]


class LAMLColumn:
    """Basic class for pair - column, role."""

    def __init__(self, data: Any, role: ColumnRole):
        """Set a pair column/role.

        Args:
            data: 1d array like.
            role: Column role.

        """
        self.data = data
        self.role = role

    def __repr__(self) -> str:
        """Repr method.

        Returns:
            String with data representation.

        """
        return self.data.__repr__()


class LAMLDataset:
    """Basic class to create dataset."""

    # TODO: Create checks here
    _init_checks = ()  # list of functions that checks that _array_like_attrs are valid
    _data_checks = ()  # list of functions that checks that data in .set_data is valid for _array_like_attrs
    _concat_checks = ()  # list of functions that checks that datasets for concatenation are valid
    _dataset_type = "LAMLDataset"

    def __init__(
        self,
        data: Any,
        features: Optional[list],
        roles: Optional[RolesDict],
        task: Optional[Task] = None,
        **kwargs: Any
    ):
        """Create dataset with given data, features, roles and special attributes.

        Args:
            data: 2d array of data of special type for each dataset type.
            features: Feature names or None for empty data.
            roles: Features roles or None for empty data.
            task: Task for dataset if train/valid.
            **kwargs: Special named array of attributes (target, group etc.).

        """
        if features is None:
            features = []
        if roles is None:
            roles = {}
        self._initialize(task, **kwargs)
        if data is not None:
            self.set_data(data, features, roles)

    def __len__(self):
        """Get count of rows in dataset.

        Returns:
            Number of rows in dataset.

        """
        return self.shape[0]

    def __repr__(self):
        """Get str representation.

        Returns:
            String with data representation.

        """
        # TODO: View for empty
        return self.data.__repr__()

    # default behavior and abstract methods
    def __getitem__(self, k: Tuple[RowSlice, ColSlice]) -> Union["LAMLDataset", LAMLColumn]:
        """Select a subset of dataset.

        Define how to slice a dataset
        in way ``dataset[[1, 2, 3...], ['feat_0', 'feat_1'...]]``.
        Default behavior based on ``._get_cols``, ``._get_rows``, ``._get_2d``.

        Args:
            k: First element optional integer columns indexes,
                second - optional feature name or list of features names.

        Returns:
            Subdataset.

        """
        # TODO: Maybe refactor this part?
        if type(k) is tuple:
            rows, cols = k
        else:
            rows = k
            cols = None

        # case when columns are defined
        if cols is not None:
            idx = self._get_cols_idx(cols)
            data = self._get_2d(self.data, (rows, idx))

            # case of single column - return LAMLColumn
            if isinstance(cols, str):
                dataset = LAMLColumn(self._get_2d(self.data, (rows, idx)), role=self.roles[cols])

                return dataset

            # case of multiple columns - return LAMLDataset
            roles = dict(((x, self.roles[x]) for x in self.roles if x in cols))
            features = [x for x in cols if x in set(self.features)]
        else:
            data, roles, features = self.data, self.roles, self.features

        # case when rows are defined
        if rows is None:
            dataset = self.empty()
        else:
            dataset = copy(self)
            params = dict(((x, self._get_rows(self.__dict__[x], rows)) for x in self._array_like_attrs))
            dataset._initialize(self.task, **params)
            data = self._get_rows(data, rows)

        dataset.set_data(data, features, roles)

        return dataset

    def __setitem__(self, k: str, val: Any):
        """Inplace set values for single column (in default implementation).

        Args:
            k: Feature name.
            val: :class:`~lightautoml.dataset.base.LAMLColumn`
                or 1d array like.

        """
        assert k in self.features, "Can only replace existed columns in default implementations."
        idx = self._get_cols_idx(k)
        # for case when setting col and change role
        if type(val) is LAMLColumn:
            assert val.role.dtype == self.roles[k].dtype, "Inplace changing types unavailable."
            self._set_col(self.data, idx, val.data)
            self.roles[k] = val.role
        # for case only changing column values
        else:
            self._set_col(self.data, idx, val)

    def __getattr__(self, item: str) -> Any:
        """Get item for key features as target/folds/weights etc.

        Args:
            item: Attribute name.

        Returns:
            Attribute value.

        """
        if item in valid_array_attributes:
            return None
        raise AttributeError

    @property
    def features(self) -> list:
        """Define how to get features names list.

        Returns:
            Features names.

        """
        return list(self._features)

    @features.setter
    def features(self, val: list):
        """Define how to set features list.

        Args:
            val: Features names.

        """
        self._features = copy(val)

    @property
    def data(self) -> Any:
        """Get data attribute.

        Returns:
            Any, array like or ``None``.

        """
        return self._data

    @data.setter
    def data(self, val: Any):
        """Set data array or ``None``.

        Args:
            val: Some data or ``None``.

        """
        self._data = val

    @property
    def roles(self) -> RolesDict:
        """Get roles dict.

        Returns:
            Dict of feature roles.

        """
        return copy(self._roles)

    @roles.setter
    def roles(self, val: RolesDict):
        """Set roles dict.

        Args:
            val: Roles dict.

        """
        self._roles = dict(((x, val[x]) for x in self.features))

    @property
    def inverse_roles(self) -> Dict[ColumnRole, List[str]]:
        """Get inverse dict of feature roles.

        Returns:
            dict, keys - roles, values - features names.

        """
        inv_roles = {}

        roles = self.roles

        for k in roles:
            r = roles[k]
            if r in inv_roles:
                inv_roles[r].append(k)
            else:
                inv_roles[r] = [k]

        return inv_roles

    def _initialize(self, task: Optional[Task], **kwargs: Any):
        """Initialize empty dataset with task and array like attributes.

        Args:
            task: Task name for dataset.
            **kwargs: 1d arrays like attrs like target, group etc.

        """
        assert all([x in valid_array_attributes for x in kwargs]), "Unknown array attribute. Valid are {0}".format(
            valid_array_attributes
        )

        self.task = task
        # here we set target and group and so ...
        self._array_like_attrs = []
        for k in kwargs:
            self._array_like_attrs.append(k)
            self.__dict__[k] = kwargs[k]

        # checks for valid values in target, groups ...
        for check in self._init_checks:
            check(self)

        # set empty attributes
        self._data = None
        self._features = []
        self._roles = {}

    def set_data(self, data: Any, features: Any, roles: Any):
        """Inplace set data, features, roles for empty dataset.

        Args:
            data: 2d array like or ``None``.
            features: List of features names.
            roles: Roles dict.

        """
        self.data = data
        self.features = features
        self.roles = roles

        # data checks
        for check in self._data_checks:
            check(self)

    def empty(self) -> "LAMLDataset":
        """Get new dataset for same task and targets, groups, without features.

        Returns:
            New empty dataset.

        """
        dataset = copy(self)
        params = dict(((x, self.__dict__[x]) for x in self._array_like_attrs))
        dataset._initialize(self.task, **params)

        return dataset

    def _get_cols_idx(self, columns: Sequence) -> Union[List[int], int]:
        """Get numeric index of columns by column names.

        Args:
            columns: Features names.

        Returns:
            List of integer indexes of single int.

        """
        if isinstance(columns, str):
            idx = self.features.index(columns)

        else:
            idx = [self.features.index(x) for x in columns]

        return idx

    # default calculated properties
    @property
    def shape(self) -> Tuple[Optional[int], Optional[int]]:
        """Get size of 2d feature matrix.

        Returns:
            Tuple of 2 elements.

        """
        rows, cols = None, None
        try:
            rows, cols = len(self.data), len(self.features)
        except TypeError:
            if len(self._array_like_attrs) > 0:
                rows = len(self.__dict__[self._array_like_attrs[0]])
        return rows, cols

    # static methods - how to make 1d slice, 2s slice, concat of feature matrix etc ...
    @staticmethod
    def _hstack(datasets: Sequence[Any]) -> Any:
        """Abstract method - define horizontal stack of feature arrays.

        Args:
            datasets: Sequence of feature arrays.

        Returns:  # noqa DAR202
            Single feature array.

        """
        raise NotImplementedError("Horizontal Stack not implemented.")

    @staticmethod
    def _get_rows(data, k: IntIdx) -> Any:
        """Abstract - define how to make rows slice of feature array.

        Args:
            data: 2d feature array.
            k: Sequence of int indexes or int.

        Returns:  # noqa DAR202
            2d feature array.

        """
        raise NotImplementedError("Row Slice not Implemented.")

    @staticmethod
    def _get_cols(data, k: IntIdx) -> Any:
        """Abstract - define how to make columns slice of feature array.

        Args:
            data: 2d feature array.
            k: Sequence indexes or single index.

        Returns:  # noqa DAR202
            2d feature array.

        """
        raise NotImplementedError("Column Slice not Implemented.")

    # TODO: remove classmethod here ?
    @classmethod
    def _get_2d(cls, data: Any, k: Tuple[IntIdx, IntIdx]) -> Any:
        """Default implementation of 2d slice based on rows slice and columns slice.

        Args:
            data: 2d feature array.
            k: Tuple of integer sequences or 2 int.

        Returns:
            2d feature array.

        """
        rows, cols = k

        return cls._get_rows(cls._get_cols(data, cols), rows)

    @staticmethod
    def _set_col(data: Any, k: int, val: Any):
        """Abstract - set a value of single column by column name inplace.

        Args:
            data: 2d feature array.
            k: Column idx.
            val: 1d column value.

        """
        raise NotImplementedError("Column setting inplace not implemented.")

    @classmethod
    def concat(cls, datasets: Sequence["LAMLDataset"]) -> "LAMLDataset":
        """Concat multiple dataset.

        Default behavior - takes empty dataset from datasets[0]
        and concat all features from others.

        Args:
            datasets: Sequence of datasets.

        Returns:
            Concated dataset.

        """
        for check in cls._concat_checks:
            check(datasets)

        dataset = datasets[0].empty()
        data = []
        features = []
        roles = {}

        atrs = set(dataset._array_like_attrs)
        for ds in datasets:
            data.append(ds.data)
            features.extend(ds.features)
            roles = {**roles, **ds.roles}
            for atr in ds._array_like_attrs:
                if atr not in atrs:
                    dataset._array_like_attrs.append(atr)
                    dataset.__dict__[atr] = ds.__dict__[atr]
                    atrs.update({atr})

        data = cls._hstack(data)
        dataset.set_data(data, features, roles)

        return dataset

    def drop_features(self, droplist: Sequence[str]):
        """Inplace drop columns from dataset.

        Args:
            droplist: Feature names.

        Returns:
            Dataset without columns.

        """
        if len(droplist) == 0:
            return self
        return self[:, [x for x in self.features if x not in droplist]]

    @staticmethod
    def from_dataset(dataset: "LAMLDataset") -> "LAMLDataset":
        """Abstract method - how to create this type of dataset from others.

        Args:
            dataset: Original type dataset.

        Returns:  # noqa DAR202
            Converted type dataset.

        """
        raise NotImplementedError

    @property
    def dataset_type(self):
        """Get type of dataset."""
        return self._dataset_type
